module A2D

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2D Helmholtz matrix generation  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

use NDpre2D
implicit none


contains


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  Acoustic matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine A2D_acoustic( freq, tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel, rho, A )

        complex(kind=4) :: freq, sigma0, hz, hx
        integer         :: tos, Nz, Nx, N_delta, locHZ, locHX, locTZ, locTX
        integer         :: label(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::   vel(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::   rho(locTZ-locHZ+1+2*tos,locTX-locHX+1+2*tos)
        type(CSR)       :: A

        integer                     :: zz, xx, kk, lvl, length, counter, col_tmp(9)
        complex(kind=4)             :: rx, Czz, Cxx, Czx, tmp, tmp1, tmp2, vel_tmp(9), rho_tmp(9), val_tmp(9)
        complex(kind=4),allocatable :: Sz(:), Sx(:)

        ! the ratio of the spatial step size and parameters
        rx  = hz / hx    

        ! generate Sz and Sx
        allocate(Sz(Nz+2*tos),Sx(Nx+2*tos))
        Sz = ONE
        Sx = ONE
        do counter = 1,N_delta+tos
            tmp = RT * (sigma0/freq) * ( 1.0 + cos(real(pi)*real(counter)/real(N_delta+tos)) ) / 2.0    ! here we are using a cos^2 taper function 
            Sz(counter)            = Sz(counter) - tmp
            Sz(Nz+2*tos-counter+1) = Sz(counter)
            Sx(counter)            = Sx(counter) - tmp
            Sx(Nx+2*tos-counter+1) = Sx(counter)
        end do

        ! allocate A%ind
        length = (locTZ-locHZ+1)*(locTX-locHX+1)    ! the order of the matrix A: number of unknows
        allocate(A%ind(length+1))
        A%ind = 0

        do xx = locHX,locTX
            do zz = locHZ,locTZ
               lvl = label(zz-locHZ+1,xx-locHX+1) + 1
               if (xx > locHX) then
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               end if
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if                
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then               
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               if (xx < locTX) then
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               end if
            end do
        end do

        A%ind(1) = 1
        do kk = 1,length
            A%ind(kk+1) = A%ind(kk+1) + A%ind(kk)
        end do

        ! allocate A%col and A%val
        length = 9*(locTZ-locHZ+1)*(locTX-locHX+1) - 6*((locTZ-locHZ-1) + (locTX-locHX-1)) - 20
        allocate(A%col(length),A%val(length))
        A%col = 0
        A%val = ZERO

        ! generate A
        do xx = locHX,locTX
            do zz = locHZ,locTZ

               ! col_tmp and vel_tmp
               col_tmp = 0
               vel_tmp = ZERO
               if (xx > locHX) then
                   if (zz > locHZ) then
                       col_tmp(1) = label(zz-locHZ,  xx-locHX)
                       vel_tmp(1) =   vel(zz-locHZ,  xx-locHX)
                   end if
                       col_tmp(2) = label(zz-locHZ+1,xx-locHX)
                       vel_tmp(2) =   vel(zz-locHZ+1,xx-locHX)
                   if (zz < locTZ) then
                       col_tmp(3) = label(zz-locHZ+2,xx-locHX)
                       vel_tmp(3) =   vel(zz-locHZ+2,xx-locHX)
                   end if
               end if
                   if (zz > locHZ) then
                       col_tmp(4) = label(zz-locHZ,  xx-locHX+1)
                       vel_tmp(4) =   vel(zz-locHZ,  xx-locHX+1)
                   end if                
                       col_tmp(5) = label(zz-locHZ+1,xx-locHX+1)
                       vel_tmp(5) =   vel(zz-locHZ+1,xx-locHX+1)
                   if (zz < locTZ) then               
                       col_tmp(6) = label(zz-locHZ+2,xx-locHX+1)
                       vel_tmp(6) =   vel(zz-locHZ+2,xx-locHX+1)
                   end if
               if (xx < locTX) then
                   if (zz > locHZ) then
                       col_tmp(7) = label(zz-locHZ,  xx-locHX+2)
                       vel_tmp(7) =   vel(zz-locHZ,  xx-locHX+2)
                   end if
                       col_tmp(8) = label(zz-locHZ+1,xx-locHX+2)
                       vel_tmp(8) =   vel(zz-locHZ+1,xx-locHX+2)
                   if (zz < locTZ) then
                       col_tmp(9) = label(zz-locHZ+2,xx-locHX+2)
                       vel_tmp(9) =   vel(zz-locHZ+2,xx-locHX+2)
                   end if
               end if

               ! rho_tmp: copy from rho
               rho_tmp = ZERO
               rho_tmp(1) = rho( (zz-1)-locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(2) = rho(  zz   -locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(3) = rho( (zz+1)-locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(4) = rho( (zz-1)-locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(5) = rho(  zz   -locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(6) = rho( (zz+1)-locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(7) = rho( (zz-1)-locHZ+1+tos, (xx+1)-locHX+1+tos )
               rho_tmp(8) = rho(  zz   -locHZ+1+tos, (xx+1)-locHX+1+tos )
               rho_tmp(9) = rho( (zz+1)-locHZ+1+tos, (xx+1)-locHX+1+tos )

               ! value preprocessing
               Czz = -ONE
               Cxx = -ONE
               Czx = ZERO

               ! val_tmp
               val_tmp = ZERO

               ! 1. d^2/dz^2
               tmp  = Czz * rho_tmp(5) / Sz(zz+tos)
               tmp1 = (1.0/(rho_tmp(4)*Sz(zz-1+tos))+1.0/(rho_tmp(5)*Sz(zz+tos))) / 2.0
               tmp2 = (1.0/(rho_tmp(6)*Sz(zz+1+tos))+1.0/(rho_tmp(5)*Sz(zz+tos))) / 2.0
               val_tmp(4) = val_tmp(4) + tmp * tmp1
               val_tmp(5) = val_tmp(5) - tmp *(tmp1 + tmp2)
               val_tmp(6) = val_tmp(6) + tmp * tmp2

               ! 2. d^2/dx^2
               tmp = Cxx * rx**2.0 * rho_tmp(5) / Sx(xx+tos)
               tmp1 = (1.0/(rho_tmp(2)*Sx(xx-1+tos))+1.0/(rho_tmp(5)*Sx(xx+tos))) / 2.0
               tmp2 = (1.0/(rho_tmp(8)*Sx(xx+1+tos))+1.0/(rho_tmp(5)*Sx(xx+tos))) / 2.0
               val_tmp(2) = val_tmp(2) + tmp * tmp1
               val_tmp(5) = val_tmp(5) - tmp *(tmp1 + tmp2)
               val_tmp(8) = val_tmp(8) + tmp * tmp2

               ! 3. d^2/dzdx
               tmp = Czx * rx/4.0 * rho_tmp(5) / (Sz(zz+tos)*Sx(xx+tos))
               val_tmp(1) = val_tmp(1) + tmp * (rho_tmp(2)+rho_tmp(4))/(2.0*rho_tmp(2)*rho_tmp(4))
               val_tmp(3) = val_tmp(3) - tmp * (rho_tmp(2)+rho_tmp(6))/(2.0*rho_tmp(2)*rho_tmp(6))
               val_tmp(7) = val_tmp(7) - tmp * (rho_tmp(8)+rho_tmp(4))/(2.0*rho_tmp(8)*rho_tmp(4))
               val_tmp(9) = val_tmp(9) + tmp * (rho_tmp(8)+rho_tmp(6))/(2.0*rho_tmp(8)*rho_tmp(6))

               ! 4. mass term
               val_tmp(5) = val_tmp(5) - (hz*freq/vel_tmp(5))**2.0

               ! 5. KEY step: modification for the sake of overlapping interface
               if (locHZ > 1 .and. zz == locHZ) then
                   val_tmp(2:8:3) = ZERO
               end if
               if (locHX > 1 .and. xx == locHX) then
                   val_tmp(4:6) = ZERO
               end if

               ! copy val and col
               counter = 0
               lvl = label(zz-locHZ+1,xx-locHX+1)
               do kk = 1,9
                   if (col_tmp(kk) /= 0) then
                       A%col( A%ind(lvl)+counter ) = col_tmp(kk)
                       A%val( A%ind(lvl)+counter ) = val_tmp(kk)
                       counter = counter + 1
                   end if
               end do

            end do
        end do
        deallocate(Sz,Sx)

        return
    end subroutine A2D_acoustic


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  Elliptic anisotropy  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine A2D_elliptic( freq, tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel, rho, epslon, theta, A )

        complex(kind=4) :: freq, sigma0, hz, hx
        integer         :: tos, Nz, Nx, N_delta, locHZ, locHX, locTZ, locTX
        integer         ::  label(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::    vel(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::    rho(locTZ-locHZ+1+2*tos,locTX-locHX+1+2*tos)
        real(kind=4)    :: epslon(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::  theta(locTZ-locHZ+1      ,locTX-locHX+1)
        type(CSR)       :: A

        integer                     :: zz, xx, kk, lvl, length, counter, col_tmp(9)
        complex(kind=4)             :: rx, Czz, Cxx, Czx, tmp, tmp1, tmp2, ee, tt, lambda_z, lambda_x, vel_tmp(9), rho_tmp(9), val_tmp(9)
        complex(kind=4),allocatable :: Sz(:), Sx(:)

        ! the ratio of the spatial step size and parameters
        rx  = hz / hx    

        ! generate Sz and Sx
        allocate(Sz(Nz+2*tos),Sx(Nx+2*tos))
        Sz = ONE
        Sx = ONE
        do counter = 1,N_delta+tos
            tmp = RT * (sigma0/freq) * ( 1.0 + cos(real(pi)*real(counter)/real(N_delta+tos)) ) / 2.0    ! here we are using a cos^2 taper function 
            Sz(counter)            = Sz(counter) - tmp
            Sz(Nz+2*tos-counter+1) = Sz(counter)
            Sx(counter)            = Sx(counter) - tmp
            Sx(Nx+2*tos-counter+1) = Sx(counter)
        end do

        ! allocate A%ind
        length = (locTZ-locHZ+1)*(locTX-locHX+1)    ! the order of the matrix A: number of unknows
        allocate(A%ind(length+1))
        A%ind = 0

        do xx = locHX,locTX
            do zz = locHZ,locTZ
               lvl = label(zz-locHZ+1,xx-locHX+1) + 1
               if (xx > locHX) then
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               end if
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if                
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then               
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               if (xx < locTX) then
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               end if
            end do
        end do

        A%ind(1) = 1
        do kk = 1,length
            A%ind(kk+1) = A%ind(kk+1) + A%ind(kk)
        end do

        ! allocate A%col and A%val
        length = 9*(locTZ-locHZ+1)*(locTX-locHX+1) - 6*((locTZ-locHZ-1) + (locTX-locHX-1)) - 20
        allocate(A%col(length),A%val(length))
        A%col = 0
        A%val = ZERO

        ! generate A
        do xx = locHX,locTX
            do zz = locHZ,locTZ

               ! col_tmp and vel_tmp
               col_tmp = 0
               vel_tmp = ZERO
               if (xx > locHX) then
                   if (zz > locHZ) then
                       col_tmp(1) = label(zz-locHZ,  xx-locHX)
                       vel_tmp(1) =   vel(zz-locHZ,  xx-locHX)
                   end if
                       col_tmp(2) = label(zz-locHZ+1,xx-locHX)
                       vel_tmp(2) =   vel(zz-locHZ+1,xx-locHX)
                   if (zz < locTZ) then
                       col_tmp(3) = label(zz-locHZ+2,xx-locHX)
                       vel_tmp(3) =   vel(zz-locHZ+2,xx-locHX)
                   end if
               end if
                   if (zz > locHZ) then
                       col_tmp(4) = label(zz-locHZ,  xx-locHX+1)
                       vel_tmp(4) =   vel(zz-locHZ,  xx-locHX+1)
                   end if                
                       col_tmp(5) = label(zz-locHZ+1,xx-locHX+1)
                       vel_tmp(5) =   vel(zz-locHZ+1,xx-locHX+1)
                   if (zz < locTZ) then               
                       col_tmp(6) = label(zz-locHZ+2,xx-locHX+1)
                       vel_tmp(6) =   vel(zz-locHZ+2,xx-locHX+1)
                   end if
               if (xx < locTX) then
                   if (zz > locHZ) then
                       col_tmp(7) = label(zz-locHZ,  xx-locHX+2)
                       vel_tmp(7) =   vel(zz-locHZ,  xx-locHX+2)
                   end if
                       col_tmp(8) = label(zz-locHZ+1,xx-locHX+2)
                       vel_tmp(8) =   vel(zz-locHZ+1,xx-locHX+2)
                   if (zz < locTZ) then
                       col_tmp(9) = label(zz-locHZ+2,xx-locHX+2)
                       vel_tmp(9) =   vel(zz-locHZ+2,xx-locHX+2)
                   end if
               end if

               ! rho_tmp: copy from rho
               rho_tmp = ZERO
               rho_tmp(1) = rho( (zz-1)-locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(2) = rho(  zz   -locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(3) = rho( (zz+1)-locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(4) = rho( (zz-1)-locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(5) = rho(  zz   -locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(6) = rho( (zz+1)-locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(7) = rho( (zz-1)-locHZ+1+tos, (xx+1)-locHX+1+tos )
               rho_tmp(8) = rho(  zz   -locHZ+1+tos, (xx+1)-locHX+1+tos )
               rho_tmp(9) = rho( (zz+1)-locHZ+1+tos, (xx+1)-locHX+1+tos )

               ! value preprocessing
               ee = epslon(zz-locHZ+1,xx-locHX+1)
               tt =  theta(zz-locHZ+1,xx-locHX+1) * pi/180.0
               lambda_z = -ONE
               lambda_x = -ONE - 2.0*ee
               Czz = lambda_z * cos(tt)**2.0 + lambda_x * sin(tt)**2.0
               Cxx = lambda_z * sin(tt)**2.0 + lambda_x * cos(tt)**2.0
               Czx = (lambda_z - lambda_x) * sin(2.0*tt)

               ! val_tmp
               val_tmp = ZERO

               ! 1. d^2/dz^2
               tmp  = Czz * rho_tmp(5) / Sz(zz+tos)
               tmp1 = (1.0/(rho_tmp(4)*Sz(zz-1+tos))+1.0/(rho_tmp(5)*Sz(zz+tos))) / 2.0
               tmp2 = (1.0/(rho_tmp(6)*Sz(zz+1+tos))+1.0/(rho_tmp(5)*Sz(zz+tos))) / 2.0
               val_tmp(4) = val_tmp(4) + tmp * tmp1
               val_tmp(5) = val_tmp(5) - tmp *(tmp1 + tmp2)
               val_tmp(6) = val_tmp(6) + tmp * tmp2

               ! 2. d^2/dx^2
               tmp = Cxx * rx**2.0 * rho_tmp(5) / Sx(xx+tos)
               tmp1 = (1.0/(rho_tmp(2)*Sx(xx-1+tos))+1.0/(rho_tmp(5)*Sx(xx+tos))) / 2.0
               tmp2 = (1.0/(rho_tmp(8)*Sx(xx+1+tos))+1.0/(rho_tmp(5)*Sx(xx+tos))) / 2.0
               val_tmp(2) = val_tmp(2) + tmp * tmp1
               val_tmp(5) = val_tmp(5) - tmp *(tmp1 + tmp2)
               val_tmp(8) = val_tmp(8) + tmp * tmp2

               ! 3. d^2/dzdx
               tmp = Czx * rx/4.0 * rho_tmp(5) / (Sz(zz+tos)*Sx(xx+tos))
               val_tmp(1) = val_tmp(1) + tmp * (rho_tmp(2)+rho_tmp(4))/(2.0*rho_tmp(2)*rho_tmp(4))
               val_tmp(3) = val_tmp(3) - tmp * (rho_tmp(2)+rho_tmp(6))/(2.0*rho_tmp(2)*rho_tmp(6))
               val_tmp(7) = val_tmp(7) - tmp * (rho_tmp(8)+rho_tmp(4))/(2.0*rho_tmp(8)*rho_tmp(4))
               val_tmp(9) = val_tmp(9) + tmp * (rho_tmp(8)+rho_tmp(6))/(2.0*rho_tmp(8)*rho_tmp(6))

               ! 4. mass term
               val_tmp(5) = val_tmp(5) - (hz*freq/vel_tmp(5))**2.0

               ! 5. KEY step: modification for the sake of overlapping interface
               if (locHZ > 1 .and. zz == locHZ) then
                   val_tmp(2:8:3) = ZERO
               end if
               if (locHX > 1 .and. xx == locHX) then
                   val_tmp(4:6) = ZERO
               end if

               ! copy val and col
               counter = 0
               lvl = label(zz-locHZ+1,xx-locHX+1)
               do kk = 1,9
                   if (col_tmp(kk) /= 0) then
                       A%col( A%ind(lvl)+counter ) = col_tmp(kk)
                       A%val( A%ind(lvl)+counter ) = val_tmp(kk)
                       counter = counter + 1
                   end if
               end do

            end do
        end do
        deallocate(Sz,Sx)

        return
    end subroutine A2D_elliptic


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  VTI  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine A2D_VTI( freq, tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel, rho, epslon, delta, A )

        complex(kind=4) :: freq, sigma0, hz, hx
        integer         :: tos, Nz, Nx, N_delta, locHZ, locHX, locTZ, locTX
        integer         ::  label(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::    vel(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::    rho(locTZ-locHZ+1+2*tos,locTX-locHX+1+2*tos)
        real(kind=4)    :: epslon(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::  delta(locTZ-locHZ+1      ,locTX-locHX+1)
        type(CSR)       :: A

        integer                     :: zz, xx, kk, lvl, length, counter, col_tmp(9)
        complex(kind=4)             :: rx, Czz, Cxx, Czx, tmp, tmp1, tmp2, tmp3, ee, dd, rhoX(6), rhoZ(2), vel_tmp(9), rho_tmp(9), val_tmp(9)
        complex(kind=4),allocatable :: Sz(:), Sx(:)

        ! the ratio of the spatial step size and parameters
        rx  = hz / hx    

        ! generate Sz and Sx
        allocate(Sz(Nz+2*tos),Sx(Nx+2*tos))
        Sz = ONE
        Sx = ONE
        do counter = 1,N_delta+tos
            tmp = RT * (sigma0/freq) * ( 1.0 + cos(real(pi)*real(counter)/real(N_delta+tos)) ) / 2.0    ! here we are using a cos^2 taper function 
            Sz(counter)            = Sz(counter) - tmp
            Sz(Nz+2*tos-counter+1) = Sz(counter)
            Sx(counter)            = Sx(counter) - tmp
            Sx(Nx+2*tos-counter+1) = Sx(counter)
        end do

        ! allocate A%ind
        length = (locTZ-locHZ+1)*(locTX-locHX+1)    ! the order of the matrix A: number of unknows
        allocate(A%ind(length+1))
        A%ind = 0

        do xx = locHX,locTX
            do zz = locHZ,locTZ
               lvl = label(zz-locHZ+1,xx-locHX+1) + 1
               if (xx > locHX) then
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               end if
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if                
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then               
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               if (xx < locTX) then
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               end if
            end do
        end do

        A%ind(1) = 1
        do kk = 1,length
            A%ind(kk+1) = A%ind(kk+1) + A%ind(kk)
        end do

        ! allocate A%col and A%val
        length = 9*(locTZ-locHZ+1)*(locTX-locHX+1) - 6*((locTZ-locHZ-1) + (locTX-locHX-1)) - 20
        allocate(A%col(length),A%val(length))
        A%col = 0
        A%val = ZERO

        ! generate A
        do xx = locHX,locTX
            do zz = locHZ,locTZ

               ! col_tmp and vel_tmp
               col_tmp = 0
               vel_tmp = ZERO
               if (xx > locHX) then
                   if (zz > locHZ) then
                       col_tmp(1) = label(zz-locHZ,  xx-locHX)
                       vel_tmp(1) =   vel(zz-locHZ,  xx-locHX)
                   end if
                       col_tmp(2) = label(zz-locHZ+1,xx-locHX)
                       vel_tmp(2) =   vel(zz-locHZ+1,xx-locHX)
                   if (zz < locTZ) then
                       col_tmp(3) = label(zz-locHZ+2,xx-locHX)
                       vel_tmp(3) =   vel(zz-locHZ+2,xx-locHX)
                   end if
               end if
                   if (zz > locHZ) then
                       col_tmp(4) = label(zz-locHZ,  xx-locHX+1)
                       vel_tmp(4) =   vel(zz-locHZ,  xx-locHX+1)
                   end if                
                       col_tmp(5) = label(zz-locHZ+1,xx-locHX+1)
                       vel_tmp(5) =   vel(zz-locHZ+1,xx-locHX+1)
                   if (zz < locTZ) then               
                       col_tmp(6) = label(zz-locHZ+2,xx-locHX+1)
                       vel_tmp(6) =   vel(zz-locHZ+2,xx-locHX+1)
                   end if
               if (xx < locTX) then
                   if (zz > locHZ) then
                       col_tmp(7) = label(zz-locHZ,  xx-locHX+2)
                       vel_tmp(7) =   vel(zz-locHZ,  xx-locHX+2)
                   end if
                       col_tmp(8) = label(zz-locHZ+1,xx-locHX+2)
                       vel_tmp(8) =   vel(zz-locHZ+1,xx-locHX+2)
                   if (zz < locTZ) then
                       col_tmp(9) = label(zz-locHZ+2,xx-locHX+2)
                       vel_tmp(9) =   vel(zz-locHZ+2,xx-locHX+2)
                   end if
               end if

               ! rho_tmp: copy from rho
               rho_tmp = ZERO
               rho_tmp(1) = rho( (zz-1)-locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(2) = rho(  zz   -locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(3) = rho( (zz+1)-locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(4) = rho( (zz-1)-locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(5) = rho(  zz   -locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(6) = rho( (zz+1)-locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(7) = rho( (zz-1)-locHZ+1+tos, (xx+1)-locHX+1+tos )
               rho_tmp(8) = rho(  zz   -locHZ+1+tos, (xx+1)-locHX+1+tos )
               rho_tmp(9) = rho( (zz+1)-locHZ+1+tos, (xx+1)-locHX+1+tos )

               ! value preprocessing
               ee = epslon(zz-locHZ+1,xx-locHX+1)
               dd =  delta(zz-locHZ+1,xx-locHX+1)
               Czz = -ONE
               Cxx = -ONE - 2.0*ee
               Czx = ZERO

               ! rhoZ and rhoX preprocessing
               rhoX = ZERO
               rhoX(1) = (1.0/(rho_tmp(1)*Sx(xx-1+tos))+1.0/(rho_tmp(4)*Sx(xx+tos))) / 2.0
               rhoX(2) = (1.0/(rho_tmp(7)*Sx(xx+1+tos))+1.0/(rho_tmp(4)*Sx(xx+tos))) / 2.0
               rhoX(3) = (1.0/(rho_tmp(2)*Sx(xx-1+tos))+1.0/(rho_tmp(5)*Sx(xx+tos))) / 2.0
               rhoX(4) = (1.0/(rho_tmp(8)*Sx(xx+1+tos))+1.0/(rho_tmp(5)*Sx(xx+tos))) / 2.0
               rhoX(5) = (1.0/(rho_tmp(3)*Sx(xx-1+tos))+1.0/(rho_tmp(6)*Sx(xx+tos))) / 2.0
               rhoX(6) = (1.0/(rho_tmp(9)*Sx(xx+1+tos))+1.0/(rho_tmp(6)*Sx(xx+tos))) / 2.0
               rhoZ = ZERO
               rhoZ(1) = (1.0/(rho_tmp(4)*Sz(zz-1+tos))+1.0/(rho_tmp(5)*Sz(zz+tos))) / 2.0
               rhoZ(2) = (1.0/(rho_tmp(6)*Sz(zz+1+tos))+1.0/(rho_tmp(5)*Sz(zz+tos))) / 2.0

               ! val_tmp
               val_tmp = ZERO

               ! 1. d^2/dz^2
               tmp  = Czz * rho_tmp(5) / Sz(zz+tos)
               val_tmp(4) = val_tmp(4) + tmp * rhoZ(1)
               val_tmp(5) = val_tmp(5) - tmp *(rhoZ(1) + rhoZ(2))
               val_tmp(6) = val_tmp(6) + tmp * rhoZ(2)

               ! 2. d^2/dx^2
               tmp = Cxx * rx**2.0 * rho_tmp(5) / Sx(xx+tos)
               val_tmp(2) = val_tmp(2) + tmp * rhoX(3)
               val_tmp(5) = val_tmp(5) - tmp *(rhoX(3) + rhoX(4))
               val_tmp(8) = val_tmp(8) + tmp * rhoX(4)

               ! 3. d^2/dzdx
               tmp = Czx * rx/4.0 * rho_tmp(5) / (Sz(zz+tos)*Sx(xx+tos))
               val_tmp(1) = val_tmp(1) + tmp * (rho_tmp(2)+rho_tmp(4))/(2.0*rho_tmp(2)*rho_tmp(4))
               val_tmp(3) = val_tmp(3) - tmp * (rho_tmp(2)+rho_tmp(6))/(2.0*rho_tmp(2)*rho_tmp(6))
               val_tmp(7) = val_tmp(7) - tmp * (rho_tmp(8)+rho_tmp(4))/(2.0*rho_tmp(8)*rho_tmp(4))
               val_tmp(9) = val_tmp(9) + tmp * (rho_tmp(8)+rho_tmp(6))/(2.0*rho_tmp(8)*rho_tmp(6))

               ! 4. d^4/dz^2dx^2
               tmp  = -2.0*(ee-dd) * rx**2.0 * rho_tmp(5) / ( (hz*freq/vel_tmp(5))**2.0 * Sz(zz+tos) * Sx(xx+tos) )
               tmp1 = tmp * rho_tmp(4) * rhoZ(1)
               tmp2 = tmp * rho_tmp(5) * (-rhoZ(1)-rhoZ(2))
               tmp3 = tmp * rho_tmp(6) * rhoZ(2)
        
               val_tmp(1) = val_tmp(1) + tmp1 * rhoX(1)
               val_tmp(4) = val_tmp(4) - tmp1 *(rhoX(1)+rhoX(2))
               val_tmp(7) = val_tmp(7) + tmp1 * rhoX(2)
               val_tmp(2) = val_tmp(2) + tmp2 * rhoX(3)
               val_tmp(5) = val_tmp(5) - tmp2 *(rhoX(3)+rhoX(4))
               val_tmp(8) = val_tmp(8) + tmp2 * rhoX(4)
               val_tmp(3) = val_tmp(3) + tmp3 * rhoX(5)
               val_tmp(6) = val_tmp(6) - tmp3 *(rhoX(5)+rhoX(6))
               val_tmp(9) = val_tmp(9) + tmp3 * rhoX(6)

               ! 4. mass term
               val_tmp(5) = val_tmp(5) - (hz*freq/vel_tmp(5))**2.0

               ! 5. KEY step: modification for the sake of overlapping interface
               if (locHZ > 1 .and. zz == locHZ) then
                   val_tmp(2:8:3) = ZERO
               end if
               if (locHX > 1 .and. xx == locHX) then
                   val_tmp(4:6) = ZERO
               end if

               ! copy val and col
               counter = 0
               lvl = label(zz-locHZ+1,xx-locHX+1)
               do kk = 1,9
                   if (col_tmp(kk) /= 0) then
                       A%col( A%ind(lvl)+counter ) = col_tmp(kk)
                       A%val( A%ind(lvl)+counter ) = val_tmp(kk)
                       counter = counter + 1
                   end if
               end do

            end do
        end do
        deallocate(Sz,Sx)

        return
    end subroutine A2D_VTI


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  HTI  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine A2D_HTI( freq, tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel, rho, epslon, delta, A )

        complex(kind=4) :: freq, sigma0, hz, hx
        integer         :: tos, Nz, Nx, N_delta, locHZ, locHX, locTZ, locTX
        integer         ::  label(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::    vel(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::    rho(locTZ-locHZ+1+2*tos,locTX-locHX+1+2*tos)
        real(kind=4)    :: epslon(locTZ-locHZ+1      ,locTX-locHX+1)
        real(kind=4)    ::  delta(locTZ-locHZ+1      ,locTX-locHX+1)
        type(CSR)       :: A

        integer                     :: zz, xx, kk, lvl, length, counter, col_tmp(9)
        complex(kind=4)             :: rx, Czz, Cxx, Czx, tmp, tmp1, tmp2, tmp3, ee, dd, rhoX(2), rhoZ(6), vel_tmp(9), rho_tmp(9), val_tmp(9)
        complex(kind=4),allocatable :: Sz(:), Sx(:)

        ! the ratio of the spatial step size and parameters
        rx  = hz / hx    

        ! generate Sz and Sx
        allocate(Sz(Nz+2*tos),Sx(Nx+2*tos))
        Sz = ONE
        Sx = ONE
        do counter = 1,N_delta+tos
            tmp = RT * (sigma0/freq) * ( 1.0 + cos(real(pi)*real(counter)/real(N_delta+tos)) ) / 2.0    ! here we are using a cos^2 taper function 
            Sz(counter)            = Sz(counter) - tmp
            Sz(Nz+2*tos-counter+1) = Sz(counter)
            Sx(counter)            = Sx(counter) - tmp
            Sx(Nx+2*Tos-counter+1) = Sx(counter)
        end do

        ! allocate A%ind
        length = (locTZ-locHZ+1)*(locTX-locHX+1)    ! the order of the matrix A: number of unknows
        allocate(A%ind(length+1))
        A%ind = 0

        do xx = locHX,locTX
            do zz = locHZ,locTZ
               lvl = label(zz-locHZ+1,xx-locHX+1) + 1
               if (xx > locHX) then
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               end if
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if                
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then               
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               if (xx < locTX) then
                   if (zz > locHZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
                       A%ind(lvl) = A%ind(lvl) + 1
                   if (zz < locTZ) then
                       A%ind(lvl) = A%ind(lvl) + 1
                   end if
               end if
            end do
        end do

        A%ind(1) = 1
        do kk = 1,length
            A%ind(kk+1) = A%ind(kk+1) + A%ind(kk)
        end do

        ! allocate A%col and A%val
        length = 9*(locTZ-locHZ+1)*(locTX-locHX+1) - 6*((locTZ-locHZ-1) + (locTX-locHX-1)) - 20
        allocate(A%col(length),A%val(length))
        A%col = 0
        A%val = ZERO

        ! generate A
        do xx = locHX,locTX
            do zz = locHZ,locTZ

               ! col_tmp and vel_tmp
               col_tmp = 0
               vel_tmp = ZERO
               if (xx > locHX) then
                   if (zz > locHZ) then
                       col_tmp(1) = label(zz-locHZ,  xx-locHX)
                       vel_tmp(1) =   vel(zz-locHZ,  xx-locHX)
                   end if
                       col_tmp(2) = label(zz-locHZ+1,xx-locHX)
                       vel_tmp(2) =   vel(zz-locHZ+1,xx-locHX)
                   if (zz < locTZ) then
                       col_tmp(3) = label(zz-locHZ+2,xx-locHX)
                       vel_tmp(3) =   vel(zz-locHZ+2,xx-locHX)
                   end if
               end if
                   if (zz > locHZ) then
                       col_tmp(4) = label(zz-locHZ,  xx-locHX+1)
                       vel_tmp(4) =   vel(zz-locHZ,  xx-locHX+1)
                   end if                
                       col_tmp(5) = label(zz-locHZ+1,xx-locHX+1)
                       vel_tmp(5) =   vel(zz-locHZ+1,xx-locHX+1)
                   if (zz < locTZ) then               
                       col_tmp(6) = label(zz-locHZ+2,xx-locHX+1)
                       vel_tmp(6) =   vel(zz-locHZ+2,xx-locHX+1)
                   end if
               if (xx < locTX) then
                   if (zz > locHZ) then
                       col_tmp(7) = label(zz-locHZ,  xx-locHX+2)
                       vel_tmp(7) =   vel(zz-locHZ,  xx-locHX+2)
                   end if
                       col_tmp(8) = label(zz-locHZ+1,xx-locHX+2)
                       vel_tmp(8) =   vel(zz-locHZ+1,xx-locHX+2)
                   if (zz < locTZ) then
                       col_tmp(9) = label(zz-locHZ+2,xx-locHX+2)
                       vel_tmp(9) =   vel(zz-locHZ+2,xx-locHX+2)
                   end if
               end if

               ! rho_tmp: copy from rho
               rho_tmp = ZERO
               rho_tmp(1) = rho( (zz-1)-locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(2) = rho(  zz   -locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(3) = rho( (zz+1)-locHZ+1+tos, (xx-1)-locHX+1+tos )
               rho_tmp(4) = rho( (zz-1)-locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(5) = rho(  zz   -locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(6) = rho( (zz+1)-locHZ+1+tos,  xx   -locHX+1+tos )
               rho_tmp(7) = rho( (zz-1)-locHZ+1+tos, (xx+1)-locHX+1+tos )
               rho_tmp(8) = rho(  zz   -locHZ+1+tos, (xx+1)-locHX+1+tos )
               rho_tmp(9) = rho( (zz+1)-locHZ+1+tos, (xx+1)-locHX+1+tos )

               ! value preprocessing
               ee = epslon(zz-locHZ+1,xx-locHX+1)
               dd =  delta(zz-locHZ+1,xx-locHX+1)
               Czz = -ONE - 2.0*ee
               Cxx = -ONE
               Czx = ZERO

               ! rhoZ and rhoX preprocessing
               rhoZ = ZERO
               rhoZ(1) = (1.0/(rho_tmp(1)*Sz(zz-1+tos))+1.0/(rho_tmp(2)*Sz(zz+tos))) / 2.0
               rhoZ(2) = (1.0/(rho_tmp(3)*Sz(zz+1+tos))+1.0/(rho_tmp(2)*Sz(zz+tos))) / 2.0
               rhoZ(3) = (1.0/(rho_tmp(4)*Sz(zz-1+tos))+1.0/(rho_tmp(5)*Sz(zz+tos))) / 2.0
               rhoZ(4) = (1.0/(rho_tmp(6)*Sz(zz+1+tos))+1.0/(rho_tmp(5)*Sz(zz+tos))) / 2.0
               rhoZ(5) = (1.0/(rho_tmp(7)*Sz(zz-1+tos))+1.0/(rho_tmp(8)*Sz(zz+tos))) / 2.0
               rhoZ(6) = (1.0/(rho_tmp(9)*Sz(zz+1+tos))+1.0/(rho_tmp(8)*Sz(zz+tos))) / 2.0
               rhoX = ZERO
               rhoX(1) = (1.0/(rho_tmp(2)*Sx(xx-1+tos))+1.0/(rho_tmp(5)*Sx(xx+tos))) / 2.0
               rhoX(2) = (1.0/(rho_tmp(8)*Sx(xx+1+tos))+1.0/(rho_tmp(5)*Sx(xx+tos))) / 2.0

               ! val_tmp
               val_tmp = ZERO

               ! 1. d^2/dz^2
               tmp  = Czz * rho_tmp(5) / Sz(zz+tos)
               val_tmp(4) = val_tmp(4) + tmp * rhoZ(3)
               val_tmp(5) = val_tmp(5) - tmp *(rhoZ(3) + rhoZ(4))
               val_tmp(6) = val_tmp(6) + tmp * rhoZ(4)

               ! 2. d^2/dx^2
               tmp = Cxx * rx**2.0 * rho_tmp(5) / Sx(xx+tos)
               val_tmp(2) = val_tmp(2) + tmp * rhoX(1)
               val_tmp(5) = val_tmp(5) - tmp *(rhoX(1) + rhoX(2))
               val_tmp(8) = val_tmp(8) + tmp * rhoX(2)

               ! 3. d^2/dzdx
               tmp = Czx * rx/4.0 * rho_tmp(5) / (Sz(zz+tos)*Sx(xx+tos))
               val_tmp(1) = val_tmp(1) + tmp * (rho_tmp(2)+rho_tmp(4))/(2.0*rho_tmp(2)*rho_tmp(4))
               val_tmp(3) = val_tmp(3) - tmp * (rho_tmp(2)+rho_tmp(6))/(2.0*rho_tmp(2)*rho_tmp(6))
               val_tmp(7) = val_tmp(7) - tmp * (rho_tmp(8)+rho_tmp(4))/(2.0*rho_tmp(8)*rho_tmp(4))
               val_tmp(9) = val_tmp(9) + tmp * (rho_tmp(8)+rho_tmp(6))/(2.0*rho_tmp(8)*rho_tmp(6))

               ! 4. d^4/dz^2dx^2
               tmp  = -2.0*(ee-dd) * rx**2.0 * rho_tmp(5) / ( (hz*freq/vel_tmp(5))**2.0 * Sz(zz+tos) * Sx(xx+tos) )
               tmp1 = tmp * rho_tmp(2) * rhoX(1)
               tmp2 = tmp * rho_tmp(5) * (-rhoX(1)-rhoX(2))
               tmp3 = tmp * rho_tmp(8) * rhoX(2)

               val_tmp(1) = val_tmp(1) + tmp1 * rhoZ(1)
               val_tmp(2) = val_tmp(2) - tmp1 *(rhoZ(1)+rhoZ(2))
               val_tmp(3) = val_tmp(3) + tmp1 * rhoZ(2)
               val_tmp(4) = val_tmp(4) + tmp2 * rhoZ(3)
               val_tmp(5) = val_tmp(5) - tmp2 *(rhoZ(3)+rhoZ(4))
               val_tmp(6) = val_tmp(6) + tmp2 * rhoZ(4)
               val_tmp(7) = val_tmp(7) + tmp3 * rhoZ(5)
               val_tmp(8) = val_tmp(8) - tmp3 *(rhoZ(5)+rhoZ(6))
               val_tmp(9) = val_tmp(9) + tmp3 * rhoZ(6)        

               ! 4. mass term
               val_tmp(5) = val_tmp(5) - (hz*freq/vel_tmp(5))**2.0


               ! 5. KEY step: modification for the sake of overlapping interface
               if (locHZ > 1 .and. zz == locHZ) then
                   val_tmp(2:8:3) = ZERO
               end if
               if (locHX > 1 .and. xx == locHX) then
                   val_tmp(4:6) = ZERO
               end if

               ! copy val and col
               counter = 0
               lvl = label(zz-locHZ+1,xx-locHX+1)
               do kk = 1,9
                   if (col_tmp(kk) /= 0) then
                       A%col( A%ind(lvl)+counter ) = col_tmp(kk)
                       A%val( A%ind(lvl)+counter ) = val_tmp(kk)
                       counter = counter + 1
                   end if
               end do

            end do
        end do
        deallocate(Sz,Sx)

        return
    end subroutine A2D_HTI



    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  TTI  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine A2D_TTI( freq, tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, vel_sz, rho, epslon, delta, theta, A )

        complex(kind=4) :: freq, sigma0, hz, hx
        integer         :: tos, Nz, Nx, N_delta, locHZ, locHX, locTZ, locTX
        integer         ::  label(locTZ-locHZ+1       ,locTX-locHX+1)
        real(kind=4)    :: vel_pz(locTZ-locHZ+1       ,locTX-locHX+1)
        real(kind=4)    :: vel_sz(locTZ-locHZ+1       ,locTX-locHX+1)
        real(kind=4)    ::    rho(locTZ-locHZ+1+2*tos ,locTX-locHX+1+2*tos)
        real(kind=4)    :: epslon(locTZ-locHZ+1       ,locTX-locHX+1)
        real(kind=4)    ::  delta(locTZ-locHZ+1       ,locTX-locHX+1)
        real(kind=4)    ::  theta(locTZ-locHZ+1       ,locTX-locHX+1)
        type(CSR)       :: A

        integer                     :: zz, xx, ii, jj, lvl, length, counter, col_tmp(5,5)
        complex(kind=4)             :: rx, tmp, tmp1, tmp2, tmp3, tmp4, vp, vs, ee, dd, tt, D11, D12, D13, D21, D22, D23
        complex(kind=4)             :: Czz, Cxx, Czx, Czzzz, Czzzx, Czzxx, Czxxx, Cxxxx
        complex(kind=4)             :: rho_tmp(5,5), val_tmp(5,5)
        complex(kind=4),allocatable :: Sz(:), Sx(:)

        ! the ratio of the spatial step size and parameters
        rx  = hz / hx    

        ! generate Sz and Sx
        allocate(Sz(Nz+2*tos),Sx(Nx+2*tos))
        Sz = ONE
        Sx = ONE
        do counter = 1,N_delta+tos
            tmp = RT * (sigma0/freq) * ( 1.0 + cos(real(pi)*real(counter)/real(N_delta+tos)) ) / 2.0    ! here we are using a cos^2 taper function 
            Sz(counter)            = Sz(counter) - tmp
            Sz(Nz+2*tos-counter+1) = Sz(counter)
            Sx(counter)            = Sx(counter) - tmp
            Sx(Nx+2*tos-counter+1) = Sx(counter)
        end do

        ! allocate A%ind
        length = (locTZ-locHZ+1)*(locTX-locHX+1)    ! the order of the matrix A: number of unknows
        allocate(A%ind(length+1))
        A%ind = 0

        do xx = locHX,locTX
            do zz = locHZ,locTZ
               counter = 0
               lvl = label(zz-locHZ+1,xx-locHX+1) + 1
               do jj = -2,2
                   do ii = -2,2
                       if (zz+ii>=locHZ .and. zz+ii<=locTZ .and. xx+jj>=locHX .and. xx+jj<=locTX) then
                           counter = counter + 1
                       end if
                   end do
               end do
               A%ind(lvl) = counter
            end do
        end do

        A%ind(1) = 1
        do counter = 1,length
            A%ind(counter+1) = A%ind(counter+1) + A%ind(counter)
        end do

        ! allocate A%col and A%val
        ! length = 25*(locTZ-locHZ+1)*(locTX-locHX+1) - 30*((locTZ-locHZ-1) + (locTX-locHX-1)) - 4*(16+26+9)
        length = A%ind(length+1)-1
        allocate(A%col(length),A%val(length))
        A%col = 0
        A%val = ZERO

        ! generate A
        do xx = locHX,locTX
            do zz = locHZ,locTZ

               col_tmp = 0
               rho_tmp = ZERO
               do jj = -2,2
                   do ii = -2,2
                       if (zz+ii>=locHZ .and. zz+ii<=locTZ .and. xx+jj>=locHX .and. xx+jj<=locTX) then
                           col_tmp(ii+3,jj+3) = label( zz+ii-locHZ+1,     xx+jj-locHX+1 )
                       end if
                           rho_tmp(ii+3,jj+3) =   rho( zz+ii-locHZ+1+tos, xx+jj-locHX+1+tos )
                   end do
               end do

               !! parameters
               vp = vel_pz(zz-locHZ+1,xx-locHX+1)
               vs = vel_sz(zz-locHZ+1,xx-locHX+1)
               ee = epslon(zz-locHZ+1,xx-locHX+1)
               dd =  delta(zz-locHZ+1,xx-locHX+1)
               tt =  theta(zz-locHZ+1,xx-locHX+1) * pi/180.0
               D11 = cos(tt)**2.0
               D12 = sin(tt)**2.0
               D13 = sin(2.0*tt)
               D21 = sin(tt)**2.0
               D22 = cos(tt)**2.0
               D23 = -sin(2.0*tt)
             
               ! if ( zz<=N_delta+tos .or. zz+N_delta+tos>=Nz .or. xx<=N_delta+tos .or. xx+N_delta+tos>=Nx ) then
               !     tt = 0.0E+0
               ! end if
 
               Czz   =  -D11 - (1.0+2.0*ee)*D21 - (vs/vp)**2.0
               Cxx   = (-D12 - (1.0+2.0*ee)*D22 - (vs/vp)**2.0) * rx**2.0
               Czx   = (-D13 - (1.0+2.0*ee)*D23) * rx
               Czzzz =  -2.0*(ee-dd)*(vp/hz/freq)**2.0*D11*D21 - (vs/hz/freq)**2.0*(D11**2.0 + (1.0+2.0*ee)*D21**2.0 + (2.0+2.0*dd)*D11*D21)
               Cxxxx = (-2.0*(ee-dd)*(vp/hz/freq)**2.0*D12*D22 - (vs/hz/freq)**2.0*(D12**2.0 + (1.0+2.0*ee)*D22**2.0 + (2.0+2.0*dd)*D12*D22)) * rx**4.0
               Czzzx = (-2.0*(ee-dd)*(vp/hz/freq)**2.0*(D11*D23+D21*D13) - (vs/hz/freq)**2.0*(2.0*D11*D13 + (1.0+2.0*ee)*2.0*D21*D23 & 
                        + (2.0+2.0*dd)*(D11*D23+D21*D13))) * rx
               Czxxx = (-2.0*(ee-dd)*(vp/hz/freq)**2.0*(D12*D23+D22*D13) - (vs/hz/freq)**2.0*(2.0*D12*D13 + (1.0+2.0*ee)*2.0*D22*D23 & 
                        + (2.0+2.0*dd)*(D12*D23+D22*D13))) * rx**3.0
               Czzxx = (-2.0*(ee-dd)*(vp/hz/freq)**2.0*(D11*D22+D21*D12+D13*D23) &
                        - (vs/hz/freq)**2.0*((2.0*D11*D12+D13**2.0) + (1.0+2.0*ee)*(2.0*D21*D22+D23**2.0) + (2.0+2.0*dd)*(D11*D22+D21*D12+D13*D23))) * rx**2.0

               !! val_tmp initialization
               val_tmp = ZERO


               !!!!!!!!!!!!!!!!!!!!!!!!!  1. z^2  !!!!!!!!!!!!!!!!!!!!!!!!!!!!
               tmp = Czz * rho_tmp(3,3) / Sz(zz+tos)
               tmp1 = (1.0/(rho_tmp(2,3)*Sz(zz-1+tos)) + 1.0/(rho_tmp(3,3)*Sz(zz+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(4,3)*Sz(zz+1+tos)) + 1.0/(rho_tmp(3,3)*Sz(zz+tos)))/2.0
               val_tmp(2,3) = val_tmp(2,3) + tmp * tmp1
               val_tmp(3,3) = val_tmp(3,3) - tmp *(tmp1 + tmp2)
               val_tmp(4,3) = val_tmp(4,3) + tmp * tmp2
        
        
               !!!!!!!!!!!!!!!!!!!!!!!!  2. x^2  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
               tmp = Cxx * rho_tmp(3,3) / Sx(xx+tos)
               tmp1 = (1.0/(rho_tmp(3,2)*Sx(xx-1+tos)) + 1.0/(rho_tmp(3,3)*Sx(xx+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(3,4)*Sx(xx+1+tos)) + 1.0/(rho_tmp(3,3)*Sx(xx+tos)))/2.0
               val_tmp(3,2) = val_tmp(3,2) + tmp * tmp1
               val_tmp(3,3) = val_tmp(3,3) - tmp *(tmp1 + tmp2)
               val_tmp(3,4) = val_tmp(3,4) + tmp * tmp2
        

               !!!!!!!!!!!!!!!!!!!!!!!!!!  3. zx   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
               tmp = Czx /4.0 * rho_tmp(3,3) / (Sz(zz+tos)*Sx(xx+tos))
               val_tmp(2,2) = val_tmp(2,2) + tmp * (rho_tmp(3,2)+rho_tmp(2,3))/(2.0*rho_tmp(3,2)*rho_tmp(2,3))
               val_tmp(4,2) = val_tmp(4,2) - tmp * (rho_tmp(3,2)+rho_tmp(4,3))/(2.0*rho_tmp(3,2)*rho_tmp(4,3))
               val_tmp(2,4) = val_tmp(2,4) - tmp * (rho_tmp(2,3)+rho_tmp(3,4))/(2.0*rho_tmp(2,3)*rho_tmp(3,4))
               val_tmp(4,4) = val_tmp(4,4) + tmp * (rho_tmp(4,3)+rho_tmp(3,4))/(2.0*rho_tmp(4,3)*rho_tmp(3,4))
        
        
               !!!!!!!!!!!!!!!!!!!!!!!!!!  4. z^4  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
               tmp = Czzzz * rho_tmp(3,3)/Sz(zz+tos)
               tmp1 = (1.0/(rho_tmp(1,3)*Sz(zz-2+tos)) + 1.0/(rho_tmp(2,3)*Sz(zz-1+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(2,3)*Sz(zz-1+tos)) + 1.0/(rho_tmp(3,3)*Sz(zz  +tos)))/2.0
               tmp3 = (1.0/(rho_tmp(3,3)*Sz(zz  +tos)) + 1.0/(rho_tmp(4,3)*Sz(zz+1+tos)))/2.0
               tmp4 = (1.0/(rho_tmp(4,3)*Sz(zz+1+tos)) + 1.0/(rho_tmp(5,3)*Sz(zz+2+tos)))/2.0
               val_tmp(1,3) = val_tmp(1,3) + tmp * (rho_tmp(2,3)/Sz(zz-1+tos)) * tmp1 * tmp2
               val_tmp(2,3) = val_tmp(2,3) - tmp * tmp2 * ( (rho_tmp(2,3)/Sz(zz-1+tos))*(tmp1+tmp2) + (rho_tmp(3,3)/Sz(zz+tos))*(tmp2+tmp3) )
               val_tmp(3,3) = val_tmp(3,3) + tmp * ((rho_tmp(2,3)/Sz(zz-1+tos))*tmp2**2.0 + &
                                                    (rho_tmp(3,3)/Sz(zz+tos))*(tmp2+tmp3)**2.0 + (rho_tmp(4,3)/Sz(zz+1+tos))*tmp3**2.0)
               val_tmp(4,3) = val_tmp(4,3) - tmp * tmp3 * ( (rho_tmp(4,3)/Sz(zz+1+tos))*(tmp3+tmp4) + (rho_tmp(3,3)/Sz(zz+tos))*(tmp2+tmp3) )
               val_tmp(5,3) = val_tmp(5,3) + tmp * (rho_tmp(4,3)/Sz(zz+1+tos)) * tmp3 * tmp4
        
        
               !!!!!!!!!!!!!!!!!!!!!!!!!!  5. x^4  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
               tmp = Cxxxx * rho_tmp(3,3)/Sx(xx+tos)
               tmp1 = (1.0/(rho_tmp(3,1)*Sx(xx-2+tos)) + 1.0/(rho_tmp(3,2)*Sx(xx-1+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(3,2)*Sx(xx-1+tos)) + 1.0/(rho_tmp(3,3)*Sx(xx  +tos)))/2.0
               tmp3 = (1.0/(rho_tmp(3,3)*Sx(xx  +tos)) + 1.0/(rho_tmp(3,4)*Sx(xx+1+tos)))/2.0
               tmp4 = (1.0/(rho_tmp(3,4)*Sx(xx+1+tos)) + 1.0/(rho_tmp(3,5)*Sx(xx+2+tos)))/2.0
               val_tmp(3,1) = val_tmp(3,1) + tmp * (rho_tmp(3,2)/Sx(xx-1+tos)) * tmp1 * tmp2
               val_tmp(3,2) = val_tmp(3,2) - tmp * tmp2 * ( (rho_tmp(3,2)/Sx(xx-1+tos))*(tmp1+tmp2) + (rho_tmp(3,3)/Sx(xx+tos))*(tmp2+tmp3) )
               val_tmp(3,3) = val_tmp(3,3) + tmp * ((rho_tmp(3,2)/Sx(xx-1+tos))*tmp2**2.0 + &
                                                    (rho_tmp(3,3)/Sx(xx+tos))*(tmp2+tmp3)**2.0 + (rho_tmp(3,4)/Sx(xx+1+tos))*tmp3**2.0)
               val_tmp(3,4) = val_tmp(3,4) - tmp * tmp3 * ( (rho_tmp(3,4)/Sx(xx+1+tos))*(tmp3+tmp4) + (rho_tmp(3,3)/Sx(xx+tos))*(tmp2+tmp3) )
               val_tmp(3,5) = val_tmp(3,5) + tmp * (rho_tmp(3,4)/Sx(xx+1+tos)) * tmp3 * tmp4
        
        
               !!!!!!!!!!!!!!!!!!!!!!!!!  6. z^3 x  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
               tmp =  Czzzx * rho_tmp(2,2)/Sz(zz-1+tos) * rho_tmp(3,3)/(4.0*Sz(zz+tos)*Sx(xx+tos))*(rho_tmp(3,2)+rho_tmp(2,3))/(2.0*rho_tmp(3,2)*rho_tmp(2,3))
               tmp1 = (1.0/(rho_tmp(1,2)*Sz(zz-2+tos)) + 1.0/(rho_tmp(2,2)*Sz(zz-1+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(3,2)*Sz(zz  +tos)) + 1.0/(rho_tmp(2,2)*Sz(zz-1+tos)))/2.0
               val_tmp(1,2) = val_tmp(1,2) + tmp * tmp1
               val_tmp(2,2) = val_tmp(2,2) - tmp *(tmp1 + tmp2)
               val_tmp(3,2) = val_tmp(3,2) + tmp * tmp2
        
               tmp = -Czzzx * rho_tmp(4,2)/Sz(zz+1+tos) * rho_tmp(3,3)/(4.0*Sz(zz+tos)*Sx(xx+tos))*(rho_tmp(3,2)+rho_tmp(4,3))/(2.0*rho_tmp(3,2)*rho_tmp(4,3))
               tmp1 = (1.0/(rho_tmp(3,2)*Sz(zz  +tos)) + 1.0/(rho_tmp(4,2)*Sz(zz+1+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(5,2)*Sz(zz+2+tos)) + 1.0/(rho_tmp(4,2)*Sz(zz+1+tos)))/2.0
               val_tmp(3,2) = val_tmp(3,2) + tmp * tmp1
               val_tmp(4,2) = val_tmp(4,2) - tmp *(tmp1 + tmp2)
               val_tmp(5,2) = val_tmp(5,2) + tmp * tmp2
        
               tmp = -Czzzx * rho_tmp(2,4)/Sz(zz-1+tos) * rho_tmp(3,3)/(4.0*Sz(zz+tos)*Sx(xx+tos))*(rho_tmp(2,3)+rho_tmp(3,4))/(2.0*rho_tmp(2,3)*rho_tmp(3,4))
               tmp1 = (1.0/(rho_tmp(1,4)*Sz(zz-2+tos)) + 1.0/(rho_tmp(2,4)*Sz(zz-1+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(3,4)*Sz(zz  +tos)) + 1.0/(rho_tmp(2,4)*Sz(zz-1+tos)))/2.0
               val_tmp(1,4) = val_tmp(1,4) + tmp * tmp1
               val_tmp(2,4) = val_tmp(2,4) - tmp *(tmp1 + tmp2)
               val_tmp(3,4) = val_tmp(3,4) + tmp * tmp2
        
               tmp =  Czzzx * rho_tmp(4,4)/Sz(zz+1+tos) * rho_tmp(3,3)/(4.0*Sz(zz+tos)*Sx(xx+tos))*(rho_tmp(4,3)+rho_tmp(3,4))/(2.0*rho_tmp(4,3)*rho_tmp(3,4))
               tmp1 = (1.0/(rho_tmp(3,4)*Sz(zz  +tos)) + 1.0/(rho_tmp(4,4)*Sz(zz+1+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(5,4)*Sz(zz+2+tos)) + 1.0/(rho_tmp(4,4)*Sz(zz+1+tos)))/2.0
               val_tmp(3,4) = val_tmp(3,4) + tmp * tmp1
               val_tmp(4,4) = val_tmp(4,4) - tmp *(tmp1 + tmp2)
               val_tmp(5,4) = val_tmp(5,4) + tmp * tmp2
        
        
               !!!!!!!!!!!!!!!!!!!!!!!!!!!!  7. z x^3  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
               tmp =  Czxxx * rho_tmp(2,2)/Sx(xx-1+tos) * rho_tmp(3,3)/(4.0*Sz(zz+tos)*Sx(xx+tos))*(rho_tmp(3,2)+rho_tmp(2,3))/(2.0*rho_tmp(3,2)*rho_tmp(2,3))
               tmp1 = (1.0/(rho_tmp(2,1)*Sx(xx-2+tos)) + 1.0/(rho_tmp(2,2)*Sx(xx-1+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(2,3)*Sx(xx  +tos)) + 1.0/(rho_tmp(2,2)*Sx(xx-1+tos)))/2.0
               val_tmp(2,1) = val_tmp(2,1) + tmp * tmp1
               val_tmp(2,2) = val_tmp(2,2) - tmp *(tmp1 + tmp2)
               val_tmp(2,3) = val_tmp(2,3) + tmp * tmp2
        
               tmp = -Czxxx * rho_tmp(2,4)/Sx(xx+1+tos) * rho_tmp(3,3)/(4.0*Sz(zz+tos)*Sx(xx+tos))*(rho_tmp(2,3)+rho_tmp(3,4))/(2.0*rho_tmp(2,3)*rho_tmp(3,4))
               tmp1 = (1.0/(rho_tmp(2,3)*Sx(xx  +tos)) + 1.0/(rho_tmp(2,4)*Sx(xx+1+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(2,5)*Sx(xx+2+tos)) + 1.0/(rho_tmp(2,4)*Sx(xx+1+tos)))/2.0
               val_tmp(2,3) = val_tmp(2,3) + tmp * tmp1
               val_tmp(2,4) = val_tmp(2,4) - tmp *(tmp1 + tmp2)
               val_tmp(2,5) = val_tmp(2,5) + tmp * tmp2
        
               tmp = -Czxxx * rho_tmp(4,2)/Sx(xx-1+tos) * rho_tmp(3,3)/(4.0*Sz(zz+tos)*Sx(xx+tos))*(rho_tmp(3,2)+rho_tmp(4,3))/(2.0*rho_tmp(3,2)*rho_tmp(4,3))
               tmp1 = (1.0/(rho_tmp(4,1)*Sx(xx-2+tos)) + 1.0/(rho_tmp(4,2)*Sx(xx-1+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(4,3)*Sx(xx  +tos)) + 1.0/(rho_tmp(4,2)*Sx(xx-1+tos)))/2.0
               val_tmp(4,1) = val_tmp(4,1) + tmp * tmp1
               val_tmp(4,2) = val_tmp(4,2) - tmp *(tmp1 + tmp2)
               val_tmp(4,3) = val_tmp(4,3) + tmp * tmp2
        
               tmp =  Czxxx * rho_tmp(4,4)/Sx(xx+1+tos) * rho_tmp(3,3)/(4.0*Sz(zz+tos)*Sx(xx+tos))*(rho_tmp(4,3)+rho_tmp(3,4))/(2.0*rho_tmp(4,3)*rho_tmp(3,4))
               tmp1 = (1.0/(rho_tmp(4,3)*Sx(xx  +tos)) + 1.0/(rho_tmp(4,4)*Sx(xx+1+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(4,5)*Sx(xx+2+tos)) + 1.0/(rho_tmp(4,4)*Sx(xx+1+tos)))/2.0
               val_tmp(4,3) = val_tmp(4,3) + tmp * tmp1
               val_tmp(4,4) = val_tmp(4,4) - tmp *(tmp1 + tmp2)
               val_tmp(4,5) = val_tmp(4,5) + tmp * tmp2
        
        
               !!!!!!!!!!!!!!!!!!!!!!!!!  8. z^2 x^2 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
               tmp =  Czzxx/2.0 * rho_tmp(3,2)/Sz(zz+tos) * rho_tmp(3,3)/Sx(xx+tos)*(1.0/(rho_tmp(3,2)*Sx(xx-1+tos))+1.0/(rho_tmp(3,3)*Sx(xx+tos)))/2.0
               tmp1 = (1.0/(rho_tmp(2,2)*Sz(zz-1+tos)) + 1.0/(rho_tmp(3,2)*Sz(zz+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(4,2)*Sz(zz+1+tos)) + 1.0/(rho_tmp(3,2)*Sz(zz+tos)))/2.0
               val_tmp(2,2) = val_tmp(2,2) + tmp * tmp1
               val_tmp(3,2) = val_tmp(3,2) - tmp *(tmp1 + tmp2)
               val_tmp(4,2) = val_tmp(4,2) + tmp * tmp2
        
               tmp = -Czzxx/2.0 * rho_tmp(3,3)/Sz(zz+tos) * rho_tmp(3,3)/Sx(xx+tos) &
                     * ((1.0/(rho_tmp(3,2)*Sx(xx-1+tos))+1.0/(rho_tmp(3,3)*Sx(xx+tos)))/2.0+(1.0/(rho_tmp(3,4)*Sx(xx+1+tos)) + 1.0/(rho_tmp(3,3)*Sx(xx+tos)))/2.0)
               tmp1 = (1.0/(rho_tmp(2,3)*Sz(zz-1+tos)) + 1.0/(rho_tmp(3,3)*Sz(zz+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(4,3)*Sz(zz+1+tos)) + 1.0/(rho_tmp(3,3)*Sz(zz+tos)))/2.0
               val_tmp(2,3) = val_tmp(2,3) + tmp * tmp1
               val_tmp(3,3) = val_tmp(3,3) - tmp *(tmp1 + tmp2)
               val_tmp(4,3) = val_tmp(4,3) + tmp * tmp2
          
               tmp =  Czzxx/2.0 * rho_tmp(3,4)/Sz(zz+tos) * rho_tmp(3,3)/Sx(xx+tos)*(1.0/(rho_tmp(3,4)*Sx(xx+1+tos))+1.0/(rho_tmp(3,3)*Sx(xx+tos)))/2.0
               tmp1 = (1.0/(rho_tmp(2,4)*Sz(zz-1+tos)) + 1.0/(rho_tmp(3,4)*Sz(zz+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(4,4)*Sz(zz+1+tos)) + 1.0/(rho_tmp(3,4)*Sz(zz+tos)))/2.0
               val_tmp(2,4) = val_tmp(2,4) + tmp * tmp1
               val_tmp(3,4) = val_tmp(3,4) - tmp *(tmp1 + tmp2)
               val_tmp(4,4) = val_tmp(4,4) + tmp * tmp2
        
        
               !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  9. x^2 z^2  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
               tmp = Czzxx/2.0 * rho_tmp(2,3)/Sx(xx+tos) * rho_tmp(3,3)/Sz(zz+tos)*(1.0/(rho_tmp(2,3)*Sz(zz-1+tos))+1.0/(rho_tmp(3,3)*Sz(zz+tos)))/2.0
               tmp1 = (1.0/(rho_tmp(2,2)*Sx(xx-1+tos)) + 1.0/(rho_tmp(2,3)*Sx(xx+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(2,4)*Sx(xx+1+tos)) + 1.0/(rho_tmp(2,3)*Sx(xx+tos)))/2.0
               val_tmp(2,2) = val_tmp(2,2) + tmp * tmp1
               val_tmp(2,3) = val_tmp(2,3) - tmp *(tmp1 + tmp2)
               val_tmp(2,4) = val_tmp(2,4) + tmp * tmp2
        
               tmp = -Czzxx/2.0 * rho_tmp(3,3)/Sx(xx+tos) * rho_tmp(3,3)/Sz(zz+tos) &
                     * ((1.0/(rho_tmp(2,3)*Sz(zz-1+tos))+1.0/(rho_tmp(3,3)*Sz(zz+tos)))/2.0+(1.0/(rho_tmp(4,3)*Sz(zz+1+tos))+1.0/(rho_tmp(3,3)*Sz(zz+tos)))/2.0)
               tmp1 = (1.0/(rho_tmp(3,2)*Sx(xx-1+tos)) + 1.0/(rho_tmp(3,3)*Sx(xx+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(3,4)*Sx(xx+1+tos)) + 1.0/(rho_tmp(3,3)*Sx(xx+tos)))/2.0
               val_tmp(3,2) = val_tmp(3,2) + tmp * tmp1
               val_tmp(3,3) = val_tmp(3,3) - tmp *(tmp1 + tmp2)
               val_tmp(3,4) = val_tmp(3,4) + tmp * tmp2
          
               tmp = Czzxx/2.0 * rho_tmp(4,3)/Sx(xx+tos) * rho_tmp(3,3)/Sz(zz+tos)*(1.0/(rho_tmp(4,3)*Sz(zz+1+tos))+1.0/(rho_tmp(3,3)*Sz(zz+tos)))/2.0
               tmp1 = (1.0/(rho_tmp(4,2)*Sx(xx-1+tos)) + 1.0/(rho_tmp(4,3)*Sx(xx+tos)))/2.0
               tmp2 = (1.0/(rho_tmp(4,4)*Sx(xx+1+tos)) + 1.0/(rho_tmp(4,3)*Sx(xx+tos)))/2.0
               val_tmp(4,2) = val_tmp(4,2) + tmp * tmp1
               val_tmp(4,3) = val_tmp(4,3) - tmp *(tmp1 + tmp2)
               val_tmp(4,4) = val_tmp(4,4) + tmp * tmp2


               !!!!!!!!!!!!!!!!!!!!!!!  mass term  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
               val_tmp(3,3) = val_tmp(3,3) - (hz*freq/vp)**2.0


               ! very very careful!!!!!!!!!!!!!!!!!!!
               ! 5. KEY step: modification for the sake of overlapping interface
               if (locHZ > 1 .and. zz < locHZ+tos) then
                   val_tmp(1:3+locHZ+tos-1-zz,:) = ZERO
               end if
               if (locHX > 1 .and. xx < locHX+tos) then
                   val_tmp(:,1:3+locHX+tos-1-xx) = ZERO
               end if

               ! copy val and col
               counter = 0
               lvl = label(zz-locHZ+1,xx-locHX+1)
               do jj = 1,5
                   do ii = 1,5
                       if (col_tmp(ii,jj) /= 0) then
                           A%col( A%ind(lvl)+counter ) = col_tmp(ii,jj)
                           A%val( A%ind(lvl)+counter ) = val_tmp(ii,jj)
                           counter = counter + 1
                       end if
                   end do
               end do

            end do
        end do
        deallocate(Sz,Sx)


        return
    end subroutine A2D_TTI


end module A2D
