module NDpre2D

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!  2D nested dissection and domain decomposition  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

use partition
implicit none

    ! 2D pts structure
    type,private :: pts_struct
        integer :: headZ, tailZ, headX, tailX
    end type pts_struct

    ! label: this is local label, no necessity to store the global label
    integer,allocatable :: label(:,:)


contains

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!  the final subroutine  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine NDpre_generate(myrank,tos,Lvl,Lpar,Nz,Nx,lengbTT,lenlocTT,gbTT,locTT,gbTlabel,locTlabel,gbNB,locNB,locHZ,locTZ,locHX,locTX,label,flag)

        integer                     :: myrank, tos, Lvl, Lpar, Nz, Nx, locHZ, locTZ, locHX, locTX, lengbTT, lenlocTT, flag
        type(TT_struct),allocatable :: gbTT(:), locTT(:)
        integer,allocatable         :: gbTlabel(:), locTlabel(:), label(:,:)
        type(NB_struct),allocatable :: gbNB(:), locNB(:)

        type(TL_struct)              :: TL
        type(pts_struct),allocatable :: gbpts(:), locpts(:)

        lengbTT  = 2**Lpar-1
        lenlocTT = 2**(Lvl-Lpar+2)-1
       
        ! 1. TT: elimination tree 
        call TT_generate( lengbTT, gbTT )
        call TT_generate( lenlocTT, locTT )
        

        ! 2. gbpts generate
        call TL_generate( lengbTT, gbTT, TL )
        flag = TL%ary( TL%ind(Lpar) + myrank )
        call pts_generate( 0, 1, Nz, 1, Nx, tos, lengbTT, gbTT, TL, gbpts )
        deallocate(TL%ind,TL%ary)
       

        ! 3. locpts generate
        call TL_generate( lenlocTT, locTT, TL )
        call pts_generate( Lpar-1, gbpts(flag)%headZ, gbpts(flag)%tailZ, gbpts(flag)%headX, gbpts(flag)%tailX, tos, lenlocTT, locTT, TL, locpts )
        deallocate(TL%ind,TL%ary)
        

        ! 4. Tlabel generate
        call Tlabel_generate( lengbTT, gbpts, gbTlabel )
        call Tlabel_generate( lenlocTT, locpts, locTlabel )
       

        ! 5. label local generate
        call label_generate( Nz, Nx, tos, Lpar, flag, gbTT, gbpts, lenlocTT, locTT, locpts, locHZ, locTZ, locHX, locTX, label )
        

        ! 6. neighbor generate
        call gbNB_generate( tos, lengbTT, gbTT, gbpts, gbTlabel, gbNB )
        call locNB_generate( tos, Lpar, lengbTT, lenlocTT, gbTT, locTT, gbpts, locpts, label, locNB )
        deallocate(gbpts,locpts)

        return
    end subroutine NDpre_generate


    !!!!!!!!!!!!!!!!!!!!!!!!!  1. pts generate for both global and local  !!!!!!!!!!!!!!!!!!!!!!
    subroutine pts_generate( AL, headZ, tailZ, headX, tailX, tos, lenTT, TT, TL, pts )

        ! AL stands for "Additional Level"
        integer                      :: AL, headZ, tailZ, headX, tailX, lenTT, tos
        type(TT_struct)              :: TT(lenTT)
        type(TL_struct)              :: TL
        type(pts_struct),allocatable :: pts(:)

        integer         :: level, Lvl, Ntmp, kidL, kidR, counter

        allocate(pts(lenTT))
        ! initialize
        pts(lenTT)%headZ = headZ
        pts(lenTT)%tailZ = tailZ
        pts(lenTT)%headX = headX
        pts(lenTT)%tailX = tailX
        Lvl = TT(1)%LL - 1

        do level = 1,Lvl
            do counter = TL%ind(level),TL%ind(level+1)-1

                Ntmp = TL%ary(counter)
                kidL = TT(Ntmp)%kids(1)
                kidR = TT(Ntmp)%kids(2)

                if (mod(level+AL,2) == 1) then    ! X direction

                    pts(kidL)%headZ = pts(Ntmp)%headZ
                    pts(kidL)%tailZ = pts(Ntmp)%tailZ
                    pts(kidL)%headX = pts(Ntmp)%headX

                    pts(kidR)%headZ = pts(Ntmp)%headZ
                    pts(kidR)%tailZ = pts(Ntmp)%tailZ
                    pts(kidR)%tailX = pts(Ntmp)%tailX

                    pts(Ntmp)%headX = (pts(Ntmp)%headX + pts(Ntmp)%tailX - (tos-1)) / 2
                    pts(Ntmp)%tailX = pts(Ntmp)%headX + (tos-1)
                    pts(kidL)%tailX = pts(Ntmp)%headX - 1
                    pts(kidR)%headX = pts(Ntmp)%tailX + 1

                else    ! Z direction

                    pts(kidL)%headX = pts(Ntmp)%headX
                    pts(kidL)%tailX = pts(Ntmp)%tailX
                    pts(kidL)%headZ = pts(Ntmp)%headZ

                    pts(kidR)%headX = pts(Ntmp)%headX
                    pts(kidR)%tailX = pts(Ntmp)%tailX
                    pts(kidR)%tailZ = pts(Ntmp)%tailZ

                    pts(Ntmp)%headZ = (pts(Ntmp)%headZ + pts(Ntmp)%tailZ - (tos-1)) / 2
                    pts(Ntmp)%tailZ = pts(Ntmp)%headZ + (tos-1)
                    pts(kidL)%tailZ = pts(Ntmp)%headZ - 1
                    pts(kidR)%headZ = pts(Ntmp)%tailZ + 1

                end if

            end do
        end do
    
        return
    end subroutine pts_generate


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2. Tlabel structure  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine Tlabel_generate( lenTT, pts, Tlabel )

        integer             :: lenTT
        type(pts_struct)    :: pts(lenTT)
        integer,allocatable :: Tlabel(:)
        integer             :: Ntmp
        
        allocate(Tlabel(lenTT+1))
        Tlabel(1) = 1
        do Ntmp = 1,lenTT
            Tlabel(Ntmp+1) = Tlabel(Ntmp) + (pts(Ntmp)%tailZ-pts(Ntmp)%headZ+1)*(pts(Ntmp)%tailX-pts(Ntmp)%headX+1)
        end do
        
        return
    end subroutine Tlabel_generate


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. label generate  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine label_generate( Nz, Nx, tos, Lpar, flag, gbTT, gbpts, lenlocTT, locTT, locpts, locHZ, locTZ, locHX, locTX, label )

        integer             :: tos, Lpar, flag, lenlocTT
        type(TT_struct)     :: gbTT(*), locTT(lenlocTT)
        type(pts_struct)    :: gbpts(*)
        type(pts_struct)    :: locpts(lenlocTT)
        integer             :: Nz, Nx, locHZ, locTZ, locHX, locTX
        integer,allocatable :: label(:,:)

        integer :: Ntmp, counter, headZ, tailZ, headX, tailX, zz, xx

        ! compute the span of the current domain 
        locHZ = gbpts(flag)%headZ
        locTZ = gbpts(flag)%tailZ
        locHX = gbpts(flag)%headX
        locTX = gbpts(flag)%tailX
        headZ = locHZ
        tailZ = locTZ
        headX = locHX
        tailX = locTX
        if (locHZ >  1) then
            locHZ = locHZ - tos
        end if
        if (locTZ < Nz) then
            locTZ = locTZ + tos
        end if
        if (locHX >  1) then
            locHX = locHX - tos
        end if
        if (locTX < Nx) then
            locTX = locTX + tos
        end if
        allocate(label(locTZ-locHZ+1,locTX-locHX+1))
        label = 0
        
        ! labeling inner domain points
        counter = 0
        do Ntmp = 1,lenlocTT
            if ( mod(locTT(Ntmp)%LL+Lpar-1,2) == 1 .and. locTT(Ntmp)%kids(1) /= 0 ) then
                do zz = locpts(Ntmp)%headZ,locpts(Ntmp)%tailZ
                    do xx = locpts(Ntmp)%headX,locpts(Ntmp)%tailX
                        counter = counter + 1
                        label(zz-locHZ+1,xx-locHX+1) = counter
                    end do
                end do
            else
                do xx = locpts(Ntmp)%headX,locpts(Ntmp)%tailX
                    do zz = locpts(Ntmp)%headZ,locpts(Ntmp)%tailZ
                        counter = counter + 1
                        label(zz-locHZ+1,xx-locHX+1) = counter
                    end do
                end do
            end if
        end do

        ! labeling neighbors of the current domain
        Ntmp = gbTT(flag)%parent
        do while (Ntmp /= 0) 
           if (mod(gbTT(Ntmp)%LL,2) == 1) then    ! X direction 
               if (headX - gbpts(Ntmp)%tailX == 1) then
                   headX = headX - tos
                   do zz = headZ,tailZ
                       do xx = headX,headX+(tos-1)
                           counter = counter + 1
                           label(zz-locHZ+1,xx-locHX+1) = counter
                       end do
                   end do
               end if
               if (gbpts(Ntmp)%headX - tailX == 1) then
                   tailX = tailX + tos
                   do zz = headZ,tailZ
                       do xx = tailX-(tos-1),tailX
                           counter = counter + 1
                           label(zz-locHZ+1,xx-locHX+1) = counter
                       end do
                   end do
               end if
           else
               if (headZ - gbpts(Ntmp)%tailZ == 1) then
                   headZ = headZ - tos
                   do xx = headX,tailX
                       do zz = headZ,headZ+(tos-1)
                           counter = counter + 1
                           label(zz-locHZ+1,xx-locHX+1) = counter 
                       end do
                   end do
               end if
               if (gbpts(Ntmp)%headZ - tailZ == 1) then
                   tailZ = tailZ + tos
                   do xx = headX,tailX
                       do zz = tailZ-(tos-1),tailZ
                           counter = counter + 1
                           label(zz-locHZ+1,xx-locHX+1) = counter
                       end do
                   end do
               end if
           end if 
           Ntmp = gbTT(Ntmp)%parent
        end do

        return
    end subroutine label_generate


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. global neighbor generate  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine gbNB_generate( tos, lengbTT, gbTT, gbpts, gbTlabel, gbNB )

        integer                            :: tos, lengbTT
        type(TT_struct)                    :: gbTT(lengbTT)
        type(pts_struct)                   :: gbpts(lengbTT)
        integer                            :: gbTlabel(lengbTT+1)
        type(NB_struct),target,allocatable :: gbNB(:)

        integer                 :: Ntmp, kidL, kidR, Ptmp, counter, headZ, tailZ, headX, tailX
        type(NB_struct),pointer :: tail

        ! allocating
        allocate(gbNB(lengbTT))
        gbNB(lengbTT)%Node  = 0
        gbNB(lengbTT)%Ohead = 0
        gbNB(lengbTT)%Otail = 0
        gbNB(lengbTT)%next => null()
        tail => null()

        ! outer loop for nodes
        do Ntmp = 1,lengbTT    ! excluding the root node of the global tree

            kidL = Ntmp
            kidR = Ntmp
            do while (gbTT(kidL)%kids(1) /= 0)
                kidL = gbTT(kidL)%kids(1)
                kidR = gbTT(kidR)%kids(2)
            end do
            headZ = gbpts(kidL)%headZ
            headX = gbpts(kidL)%headX
            tailZ = gbpts(kidR)%tailZ
            tailX = gbpts(kidR)%tailX

            counter = 0
            Ptmp = gbTT(Ntmp)%parent
            do while (Ptmp /= 0)
                if (mod(gbTT(Ptmp)%LL,2) == 1) then    ! X direction
                    if (headX - gbpts(Ptmp)%tailX == 1) then
                        headX = headX - tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => gbNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = Ptmp
                        tail%Ohead = gbTlabel(Ptmp) + (headZ - gbpts(Ptmp)%headZ)*tos
                        tail%Otail = gbTlabel(Ptmp) + (tailZ - gbpts(Ptmp)%headZ)*tos + (tos-1)
                    end if
                    if (gbpts(Ptmp)%headX - tailX == 1) then
                        tailX = tailX + tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => gbNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = Ptmp
                        tail%Ohead = gbTlabel(Ptmp) + (headZ - gbpts(Ptmp)%headZ)*tos
                        tail%Otail = gbTlabel(Ptmp) + (tailZ - gbpts(Ptmp)%headZ)*tos + (tos-1)
                    end if
                else    ! Z direction
                    if (headZ - gbpts(Ptmp)%tailZ == 1) then
                        headZ = headZ - tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => gbNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = Ptmp
                        tail%Ohead = gbTlabel(Ptmp) + (headX - gbpts(Ptmp)%headX)*tos
                        tail%Otail = gbTlabel(Ptmp) + (tailX - gbpts(Ptmp)%headX)*tos + (tos-1)
                    end if
                    if (gbpts(Ptmp)%headZ - tailZ == 1) then
                        tailZ = tailZ + tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => gbNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = Ptmp
                        tail%Ohead = gbTlabel(Ptmp) + (headX - gbpts(Ptmp)%headX)*tos
                        tail%Otail = gbTlabel(Ptmp) + (tailX - gbpts(Ptmp)%headX)*tos + (tos-1)
                    end if
                end if
                Ptmp = gbTT(Ptmp)%parent
            end do

            tail => null()

        end do
              
        return
    end subroutine gbNB_generate


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5. local neighbor generate  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine locNB_generate( tos, Lpar, lengbTT, lenlocTT, gbTT, locTT, gbpts, locpts, label, locNB )

        integer                            :: tos, Lpar, lengbTT, lenlocTT, Nz, Nx
        type(TT_struct)                    :: gbTT(lengbTT), locTT(lenlocTT)
        type(pts_struct)                   :: gbpts(lengbTT), locpts(lenlocTT)
        integer,allocatable                :: label(:,:)
        type(NB_struct),target,allocatable :: locNB(:)

        integer                 :: Ntmp, kidL, kidR, Ptmp, counter, headZ, tailZ, headX, tailX
        type(NB_struct),pointer :: tail

        allocate(locNB(lenlocTT))
        tail => null()

        do Ntmp = 1,lenlocTT

            kidL = Ntmp
            kidR = Ntmp
            do while (locTT(kidL)%kids(1) /= 0)
                kidL = locTT(kidL)%kids(1)
                kidR = locTT(kidR)%kids(2)
            end do
            headZ = locpts(kidL)%headZ
            headX = locpts(kidL)%headX
            tailZ = locpts(kidR)%tailZ
            tailX = locpts(kidR)%tailX

            counter = 0

            ! search for local tree: Ptmp /= 0
            Ptmp = locTT(Ntmp)%parent
            do while (Ptmp /= 0)
                if (mod(locTT(Ptmp)%LL+Lpar-1,2) == 1) then    ! X direction
                    if (headX - locpts(Ptmp)%tailX == 1) then
                        headX = headX - tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => locNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = Ptmp
                        tail%Ohead = label(headZ-locHZ+1,headX        -locHX+1)
                        tail%Otail = label(tailZ-locHZ+1,headX+(tos-1)-locHX+1)
                    end if
                    if (locpts(Ptmp)%headX - tailX == 1) then
                        tailX = tailX + tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => locNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = Ptmp
                        tail%Ohead = label(headZ-locHZ+1,tailX-(tos-1)-locHX+1)
                        tail%Otail = label(tailZ-locHZ+1,tailX        -locHX+1)
                    end if
                else    ! Z direction
                    if (headZ - locpts(Ptmp)%tailZ == 1) then
                        headZ = headZ - tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => locNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = Ptmp
                        tail%Ohead = label(headZ        -locHZ+1,headX-locHX+1)
                        tail%Otail = label(headZ+(tos-1)-locHZ+1,tailX-locHX+1)
                    end if
                    if (locpts(Ptmp)%headZ - tailZ == 1) then
                        tailZ = tailZ + tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => locNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = Ptmp
                        tail%Ohead = label(tailZ-(tos-1)-locHZ+1,headX-locHX+1)
                        tail%Otail = label(tailZ        -locHZ+1,tailX-locHX+1)
                    end if
                end if
                Ptmp = locTT(Ptmp)%parent
            end do

            ! search for global tree: Ptmp == 0
            Ptmp = gbTT(flag)%parent
            do while (Ptmp /= 0)
                if (mod(gbTT(Ptmp)%LL,2) == 1) then    ! X direction
                    if (headX - gbpts(Ptmp)%tailX == 1) then
                        headX = headX - tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => locNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = 0
                        tail%Ohead = label(headZ-locHZ+1,1  )
                        tail%Otail = label(tailZ-locHZ+1,tos)
                    end if
                    if (gbpts(Ptmp)%headX - tailX == 1) then
                        tailX = tailX + tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => locNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = 0
                        tail%Ohead = label(headZ-locHZ+1,locTX-(tos-1)-locHX+1)
                        tail%Otail = label(tailZ-locHZ+1,locTX        -locHX+1)
                    end if
                else    ! Z direction
                    if (headZ - gbpts(Ptmp)%tailZ == 1) then
                        headZ = headZ - tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => locNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = 0
                        tail%Ohead = label(1  ,headX-locHX+1)
                        tail%Otail = label(tos,tailX-locHX+1)
                    end if
                    if (gbpts(Ptmp)%headZ - tailZ == 1) then
                        tailZ = tailZ + tos
                        counter = counter + 1
                        if (counter == 1) then
                            tail => locNB(Ntmp)
                        else
                            allocate(tail%next)
                            tail => tail%next
                        end if
                        tail%next => null()
                        tail%Node  = 0
                        tail%Ohead = label(locTZ-(tos-1)-locHZ+1,headX-locHX+1)
                        tail%Otail = label(locTZ        -locHZ+1,tailX-locHX+1)
                    end if
                end if
                Ptmp = gbTT(Ptmp)%parent
            end do

            ! nullify tail
            tail => null()

        end do


        return
    end subroutine locNB_generate
    

end module NDpre2D
