PROGRAM main

  use parser
  use global
  !use adjoint_test
 
  IMPLICIT NONE
  character(256)  :: task

  call frompar("task",task)

  task=trim(task)

  print*,'The task is ',trim(task)
  print*,' '


  select case(task)
    case('adjoint_test')
       call adjointtest_main
    case('migration')
       call migration_main
    case('modeling')
       call modeling_main
    case('inversion')
       call inversiontest_main
    case('phasetomo')
	call phasetomo_main
    case('pmltest')
	call pmltest_main
  end select
  


END PROGRAM main
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
subroutine adjointtest_main

  use parser
  use geometries
  use global
  use adjoint_test
  use fileio
  !include "mpif.h"

  IMPLICIT NONE

  integer        :: isrc, iz, ix
  integer        :: nw, nx, nz, npml, nxp, nh ,dh, narr, nsrc
  real           :: dx,dz,dw,dIdI,dvadv
  complex        :: j
  real,allocatable      :: v(:), I(:), dv(:), adv(:), array_ap(:)


  call adjtest_get_global

  call frompar("nx",nx)
  call frompar("nz",nz)
  call frompar("nw",nw)
  call frompar("npml",npml)
  call frompar("nh",nh,1)
  call frompar("narray",narr)
  call frompar("nsrc",nsrc,1)

  call frompar("dx",dx,1000.)
  call frompar("dz",dz,1000.)
  call frompar("dw",dw,0.05)
  call frompar("dh",dh,10)

  Nxp=Nx+2*Npml

  allocate(v(nxp*nz))
  allocate(dv(nxp*nz))
  allocate(adv(nxp*nz))
  allocate(I(nxp*nz*nh))
  allocate(array_ap(nxp))

  call realinb(v,Nxp*Nz,vfile)
  call getarrayap(array_ap,arrayfile,narr,nxp)

  do iz=1,Nz;
	do ix=1,Nxp;
		dv(ix+Nxp*(iz-1))=exp(-(real(ix)-60.)**2/4-(real(iz)-30.)**2/4)
		adv(ix+Nxp*(iz-1))=0.
	enddo
  enddo

  do ix=1,Nxp*Nz*Nh
     I(ix)=0.
  enddo

  do isrc = 3,Nsrc
    call adjtest_oneevent(isrc, v,dv,I,adv, nxp,nz,nw,npml,nh,dx,dz,dw,dh,array_ap)
  enddo


  deallocate(v)
  deallocate(dv)
  deallocate(adv)
  deallocate(I)
  deallocate(array_ap)


end subroutine adjointtest_main
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
subroutine inversiontest_main

  use parser
  use geometries
  use global
  use inversion
  use fileio
  !include "mpif.h"

  IMPLICIT NONE

  character(256) :: ifile,adjfile
  integer        :: isrc, iz, ix, ih
  integer        :: nw, nx, nz, npml, nxp, nh ,dh, narr, nsrc
  real           :: dx,dz,dw,JJ
  complex        :: j
  integer,allocatable   :: h(:)
  real,allocatable      :: v(:), I(:), Itot(:), dv(:), adv(:), array_ap(:)


  call inv_get_global

  call frompar("nx",nx)
  call frompar("nz",nz)
  call frompar("nw",nw)
  call frompar("npml",npml)
  call frompar("nh",nh,1)
  call frompar("narray",narr)
  call frompar("nsrc",nsrc,1)

  call frompar("dx",dx,1000.)
  call frompar("dz",dz,1000.)
  call frompar("dw",dw,0.05)
  call frompar("dh",dh,10)

  Nxp=Nx+2*Npml

  allocate(h(nh))
  allocate(v(nxp*nz))
  allocate(dv(nxp*nz))
  allocate(adv(nxp*nz))
  allocate(I(nxp*nz*nh))
  allocate(array_ap(nxp))
  allocate(Itot(nxp*nz*nh))

  call realinb(v,Nxp*Nz,vfile)
  call getarrayap(array_ap,arrayfile,narr,nxp)

  do iz=1,Nz*Nxp
	adv(iz)=0.
  enddo

  do ih=1,Nh
  	h(ih)=(-(Nh-1)/2+(ih-1))*dh
	!print*,'h(ih)=',h(ih)
  enddo

  do ix=1,Nxp*Nz*Nh
     I(ix)=0.
     Itot(ix)=0.
  enddo


  do isrc = 3,Nsrc
    call inv_one_fwd(isrc,v,I,h, nxp,nz,nw,npml,nh,dx,dz,dw,dh,array_ap)

    do ix=1,Nxp*Nz*Nh
       Itot(ix)= Itot(ix)+I(ix)
    enddo
  enddo

  ifile=trim(ifile0)//"tot"
  call realoutb(Itot,Nxp*Nz*Nh,ifile)

  do ix=1,Nxp
     do iz=1,Nz
        do ih=1,Nh
	  Itot(ix+Nxp*(iz-1)+Nxp*Nz*(ih-1)) = Itot(ix+Nxp*(iz-1)+Nxp*Nz*(ih-1))*h(ih)*dx
	enddo
     enddo
  enddo


  do isrc = 3,Nsrc
    call inv_one_adj(isrc, v,Itot,adv,h, nxp,nz,nw,npml,nh,dx,dz,dw,dh)
  enddo

  JJ=0.
  do ix=1,Nxp*Nz*Nh
     JJ=JJ+Itot(ix)**2
  enddo

  print*,'The error function is J = ',JJ

  deallocate(h)
  deallocate(v)
  deallocate(dv)
  deallocate(adv)
  deallocate(I)
  deallocate(Itot)
  deallocate(array_ap)


end subroutine inversiontest_main


!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!

subroutine phasetomo_main

  use parser
  use geometries
  use global
  use phasetomography
  use phasetomorytov
  use phaseadjoint
  use fileio
  !include "mpif.h"

  IMPLICIT NONE

  integer        :: isrc, iz, ix, iter
  integer        :: nw, nx, nz,nt, npml, nxp, narr, nsrc,nrel,niter,w0,wf
  real           :: dx,dz,dw,dIdI,dvadv,delt
  complex        :: j
  real,allocatable      :: v(:), v0(:), adv(:), dv(:)!, array_ap(:), adj_ap(:)


  call phasetomo_get_global

  call frompar("nx",nx)
  call frompar("nz",nz)
  call frompar("nw",nw)
  call frompar("nt",nt,3001)
  call frompar("npml",npml)
  !call frompar("nh",nh,1)
  call frompar("narray",narr)
  call frompar("nsrc",nsrc,1)
  call frompar("nrelativepairs",nrel,1)
  call frompar("niterations",niter,1)

  call frompar("dx",dx,1000.)
  call frompar("dz",dz,1000.)
  call frompar("dw",dw,0.05)
  call frompar("delt",delt,0.025)
  call frompar("w0",w0,2)
  call frompar("wf",wf,50)


  Nxp=Nx+2*Npml
  !!Nw=wf-w0

  allocate(v(nxp*nz))
  allocate(dv(nxp*nz))
  allocate(v0(nxp*nz))
  allocate(adv(nxp*nz))

  call realinb(v,Nxp*Nz,vfile)
  call realinb(v0,Nxp*Nz,v0file)


  print*,'The type of tomography is ',trim(tomotype)
  print*,' '
  print*,'Initial velocity model read from ',trim(v0file)

  !call makecheckerv(v,40.,100.,30.,Nxp,Nz)

  do iz=1,Nz;
	do ix=1,Nxp;
		!v(ix+Nxp*(iz-1))=7200.            + 200*exp(-(real(ix)-255.)**2/50-(real(iz)-20.)**2/50)
		!v(ix+Nxp*(iz-1))=v0(ix+Nxp*(iz-1))+ 300*exp(-(real(ix)-170.)**2/400-(real(iz)-40.)**2/400)
                !v(ix+Nxp*(iz-1))=v(ix+Nxp*(iz-1)) - 300*exp(-(real(ix)-230.)**2/400-(real(iz)-40.)**2/400)
                !v(ix+Nxp*(iz-1))=v(ix+Nxp*(iz-1)) - 70* exp(-(real(ix)-255.)**2/50-(real(iz)-30.)**2/50)
                !v(ix+Nxp*(iz-1))=v(ix+Nxp*(iz-1)) + 100*exp(-(real(ix)-275.)**2/50-(real(iz)-30.)**2/100)
		!v0(ix+Nxp*(iz-1))= 1700.*(iz-1)/(Nz-1) + 6400.
		!v(ix+Nxp*(iz-1))= 7400.
		adv(ix+Nxp*(iz-1))=0.
		!v0(ix+Nxp*(iz-1))=6600.
		dv(ix+Nxp*(iz-1))=v(ix+Nxp*(iz-1))-v0(ix+Nxp*(iz-1))
	enddo
  enddo  
  do ix=1,Nxp;
     do iz = 50,80
	!v0(ix+Nxp*(iz-1))=200
	!v0(ix+Nxp*(11-1))=.20
	!v0(ix+Nxp*(12-1))=.20
     enddo
  enddo

  open(unit=fpJ,file=TRIM(Jfile), &
    access='sequential',form='formatted')

  select case(tomotype)
  case('adjoint_test')
    call phasetomo_adjoint(nsrc, v0, dv ,adv, nxp,nz,nw,nt,narr,nrel,w0,wf,npml,dx,dz,dw,delt)
  case('correlation')
    do iter = 1,niter
      call phasetomo_test(iter,nsrc, v, v0 ,adv, nxp,nz,nw,nt,narr,nrel,w0,wf,npml,dx,dz,dw,delt)
    enddo
  case('rytov_abs')
    do iter = 1,niter
      call phasetomo_rytov(iter,nsrc, v, v0 ,adv,nxp,nz,nw,nt,narr,nrel,w0,wf,npml,dx,dz,dw,delt)
    enddo
  case('rytov_rel')
    do iter = 1,niter
      call phasetomo_rytov(iter,nsrc, v, v0 ,adv, nxp,nz,nw,nt,narr,nrel,w0,wf,npml,dx,dz,dw,delt)
    enddo
  end select

  close(fpJ)

  deallocate(v)
  deallocate(v0)
  deallocate(adv)
  !deallocate(array_ap)


end subroutine phasetomo_main
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!

subroutine migration_main

  use parser
  use geometries
  use global
  use migration
  use phasetomography
  use fileio
  !include "mpif.h"

  IMPLICIT NONE

  integer        :: isrc, iz, ix
  integer        :: nw, nx, nz, npml, nxp, narr, nsrc,w0,wf
  real           :: dx,dz,dw
  complex        :: j
  real,allocatable      :: v(:), I(:),Itot(:), array_ap(:)


  call phasetomo_get_global

  call frompar("nx",nx)
  call frompar("nz",nz)
  call frompar("nw",nw)
  call frompar("npml",npml)
  !call frompar("nh",nh,1)
  call frompar("narray",narr)
  call frompar("nsrc",nsrc,1)
  !call frompar("niterations",niter,1)

  call frompar("dx",dx,1000.)
  call frompar("dz",dz,1000.)
  call frompar("dw",dw,0.05)
  call frompar("w0",w0,2)
  call frompar("wf",wf,50)
  !call frompar("dh",dh,10)

  Nxp=Nx+2*Npml
  !!Nw=wf-w0

  allocate(v(nxp*nz))
  !allocate(v0(nxp*nz))
  !allocate(adv(nxp*nz))
  allocate(I(nxp*nz))
  allocate(Itot(nxp*nz))
  allocate(array_ap(nxp))


  call realinb(v,Nxp*Nz,vfile)
  call getarrayap(array_ap,rappfile,narr,nxp)


  do iz=1,Nz;
	do ix=1,Nxp;
		!v(ix+Nxp*(iz-1))=6000.            + 200*exp(-(real(ix)-55.)**2/50-(real(iz)-20.)**2/50)
                !v(ix+Nxp*(iz-1))=v(ix+Nxp*(iz-1)) - 150*exp(-(real(ix)-75.)**2/50-(real(iz)-20.)**2/20)
                !v(ix+Nxp*(iz-1))=v(ix+Nxp*(iz-1)) - 70* exp(-(real(ix)-55.)**2/50-(real(iz)-30.)**2/50)
                !v(ix+Nxp*(iz-1))=v(ix+Nxp*(iz-1)) + 100*exp(-(real(ix)-75.)**2/50-(real(iz)-30.)**2/100)
		v(ix+Nxp*(iz-1))=7000.

	enddo
  enddo

  !call realinb(v,Nxp*Nz,vfile)

  do ix=1,Nxp*Nz
     Itot(ix)=0.
     I(ix)=0.
  enddo

!print*,'enter migrate'
  call migrate(nsrc, v, I, Itot, nxp,nz,nw,w0,wf,npml,dx,dz,dw,array_ap)



  deallocate(v)
  deallocate(I)
  deallocate(Itot)
  deallocate(array_ap)


end subroutine migration_main
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!

subroutine modeling_main

  use parser
  use geometries
  use global
  use modeling
  use phasetomography
  use fileio
  !include "mpif.h"

  IMPLICIT NONE

  integer        :: isrc, iz, ix
  integer        :: nw, nx, nz, nt, npml, nxp, narr, nsrc,w0,wf
  real           :: dx,dz,dw,dt
  complex        :: j
  real,allocatable      :: v(:),dv(:)!!, array_ap(:)
  character(256) :: modeltype


  call phasetomo_get_global

  call frompar("nx",nx)
  call frompar("nz",nz)
  call frompar("nw",nw)
  call frompar("npml",npml)
  call frompar("nt",nt,1028)
  !call frompar("nh",nh,1)
  call frompar("narray",narr)
  call frompar("nsrc",nsrc,1)
  !call frompar("niterations",niter,1)

  call frompar("dx",dx,1000.)
  call frompar("dz",dz,1000.)
  call frompar("dw",dw,0.05)
  call frompar("dt",dt,0.025)
  call frompar("w0",w0,2)
  call frompar("wf",wf,50)
  call frompar("modeltype",modeltype,"passive")
  !call frompar("dh",dh,10)

  Nxp=Nx+2*Npml
  !!Nw=wf-w0

  allocate(v(nxp*nz))
  allocate(dv(nxp*nz))
  !allocate(adv(nxp*nz))
  !allocate(array_ap(nxp))


  call realinb(v,Nxp*Nz,vfile)
  !!call getarrayap(array_ap,rappfile,narr,nxp)

  do ix=1,Nxp;
  	do iz=1,Nz;

		!v(ix+Nxp*(iz-1))=6000.            + 200*exp(-(real(ix)-55.)**2/50-(real(iz)-20.)**2/50)
                !v(ix+Nxp*(iz-1))=v(ix+Nxp*(iz-1)) - 150*exp(-(real(ix)-75.)**2/50-(real(iz)-20.)**2/20)
                !v(ix+Nxp*(iz-1))=v(ix+Nxp*(iz-1)) - 70* exp(-(real(ix)-55.)**2/50-(real(iz)-30.)**2/50)
                !v(ix+Nxp*(iz-1))=v(ix+Nxp*(iz-1)) + 100*exp(-(real(ix)-75.)**2/50-(real(iz)-30.)**2/100)
		!v(ix+Nxp*(iz-1))=7400.
		dv(ix+Nxp*(iz-1))=0.
	enddo

  enddo
  do ix=1,Nxp;
	!dv(ix+Nxp*(100-1))=0.6
	!dv(ix+Nxp*(240-1))=-0.4
  enddo

  call realinb(dv,Nxp*Nz,v0file)

  print*,'The type of modeling is ',trim(modeltype)
  print*,' '
  print*,'Model velocity model read from ',trim(vfile)

!print*,'enter migrate'
  select case(modeltype)
  case("passive")
    call modeldata_tele(nsrc, v,dv, nxp,nz,nw,nt,w0,wf,npml,dx,dz,dw,dt,narr)
  case("active")
    call modeldata_active(nsrc, v,dv, nxp,nz,nw,nt,w0,wf,npml,dx,dz,dw,dt,narr)
  end select



  deallocate(v)
  !!deallocate(array_ap)


end subroutine modeling_main

!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!

subroutine pmltest_main

  use parser
  use global
  use modeling
  !!use pmltest
  use fileio
  !include "mpif.h"

  IMPLICIT NONE

  integer        :: isrc, iz, ix
  integer        :: nw, nx, nz, npml, nxp,w0,wf,n1,n2
  real           :: dx,dz,dw
  complex        :: j
  real,allocatable      :: v(:)


  !!call pmltest_get_global

  call frompar("nx",nx)
  call frompar("nz",nz)
  call frompar("nw",nw)
  call frompar("npml",npml)
  call frompar("n1",n1)
  call frompar("n2",n2)

  call frompar("dx",dx,1000.)
  call frompar("dz",dz,1000.)
  call frompar("dw",dw,0.05)
  call frompar("w0",w0,2)
  call frompar("wf",wf,50)


  Nxp=Nx+2*Npml
  !!Nw=wf-w0

  allocate(v(nxp*nz))
  !call realinb(v,Nxp*Nz,vfile)

  do ix=1,Nxp;
  	do iz=1,Nz;
	  v(ix+Nxp*(iz-1))=7200.
	enddo
  enddo

  !!call pml_test(v,nxp,nz,nw,w0,wf,n1,n2,npml,dx,dz,dw)

  deallocate(v)


end subroutine pmltest_main


