program adjtest

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!  2D massively parallel HSS multifontal Helmholtz solver: HelmPar2D  !!!!!!!!!!!!!!!
!!!!!!!!!!!           Shen Wang,  Jianlin Xia,  Maarten V. de Hoop              !!!!!!!!!!!!!!!
!!!!!!!!!!!                Purdue & ConocoPhillips, June 2010                   !!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

use NDpre2D
use context
use A2D 
use fact
use solution2D
use parser
use fileio
use powernorm

implicit none
include 'mpif.h'


!! declaring variables
integer                      ::  ierr, myrank, nprocs
integer                      ::  JOB, tos, passive,winx,winz,deltax,deltaz
integer                      ::  Lvl, slvl, Lpar, rkmin, mr2d
real(kind=4)                 ::  tol, dt, dw, Jsum, Jsum1,Jsum2,Jsum3,sigmax,sigmaz
real(kind=8)                 ::  t1, t2
integer                      ::  Nz, Nx, N_delta, NRHS, Nt, Nw,Nh,lnx,lnz,Np,nsi,Ng, fps=20
 complex(kind=4)              :: sigma0, hz, hx, tmpi,tmpdi,tmpdx,tmpdz,dPhi,tmpss
integer,allocatable          ::  sourceZ(:), sourceX(:),h(:)
real(kind=4),allocatable     ::  Istack(:),Is(:),I0(:),Iadj(:),Ipr(:,:,:),dIpr(:,:,:)
real(kind=4),allocatable     ::  vel(:),vel0(:), dv(:), ga(:), gz(:),gx(:), taper(:), N(:),water(:),rayp(:)
real(kind=4),allocatable     ::  Jig(:),Jall(:)
 complex(kind=4),allocatable  ::  Y(:),Y0(:),YY(:), freq(:),Aw(:),dataw(:)
 complex(kind=4),allocatable  ::  X(:,:,:),dX(:,:,:) 
real(kind=4)                 ::  tmp, tmpz, tmpx,alpha,maxp,xis(1000000),zis(1000000),si1(1000000),si2(1000000)
integer                      ::  rhs, zz, xx, pp, iw, w0, wf,xi,zi, ig,ds,dh,ihh,hh, iss, ix, iz, headZ, headX
 character(len=256)           ::  vfile,v0file,dvfile,sfile,sfile0,rfile,rfile0,ifile,ifile0,wnum,pnum 
 !!complex(kind=4),parameter   :: pi = (3.1415926E+0,0.0E+0)
 !!complex(kind=4),parameter   :: ONE = (1.0E+0,0.0E+0), ZERO = (0.0E+0,0.0E+0), RT = (0.0E+0,1.0E+0)
real(kind=4),allocatable     ::  vel_pz(:,:), dvel_pz(:,:), rho(:,:)

!! MPI Initialization
 call MPI_Init(ierr)
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
 call MPI_Comm_size( MPI_COMM_WORLD, nprocs, ierr )

if (myrank==0) write(*,*) "nprocs is ",nprocs

!! set up variables
 call frompar('level',Lvl,14)
 call frompar('Nz',Nz,2048)
 call frompar('Nx',Nx,2048)
 !call frompar('Ns',NRHS,1)
 call frompar('Nt',Nt,2048)
 call frompar('Nw',Nw,55)
 call frompar('Nh',Nh,20)
 call frompar('dh',dh,2)
 call frompar('dx',tmp,250.); hx=tmp
 call frompar('dz',tmp,250.); hz=tmp
 call frompar('dt',dt,0.01)
 call frompar('ds',ds,10)
 call frompar('w0',w0,2)
 call frompar('wf',wf,Nw)
 call frompar('x0',xi,80)
 call frompar('z0',zi,110)
 call frompar('alpha',alpha,1.)
 call frompar('vfile',vfile)
 call frompar('v0file',v0file)
 call frompar('dvfile',dvfile)
 call frompar('sfile',sfile0)
 call frompar('rfile',rfile0,' ')
 call frompar('ifile',ifile0)
 call frompar('passive',passive,0)
 call frompar('slownesses',Np,15)
 call frompar('maxp',maxp,30.)




NRHS = Np*2 - 1


allocate(Is(Nz*Nx*NRHS*2))
allocate(Iadj(Nz*Nx*NRHS))



NRHS = NRHS * 2 !! NOTE: TO ACCOMODATE SOURCE AND RECEIVER




  Is=0.
  Iadj=0.

  N_delta = 50
  
  
   ifile=trim(ifile0)//'J'
   open( unit=20,file=TRIM(ifile), access='direct',recl=1,form='unformatted')
  
   ifile=trim(ifile0)//'J1'
   open( unit=21,file=TRIM(ifile), access='direct',recl=1,form='unformatted')   
   
   ifile=trim(ifile0)//'J2'
   open( unit=22,file=TRIM(ifile), access='direct',recl=1,form='unformatted')  
  
   ifile=trim(ifile0)//'J3'
   open( unit=23,file=TRIM(ifile), access='direct',recl=1,form='unformatted')   
  
   ifile=trim(ifile0)//'T'
   open( unit=24,file=TRIM(ifile), access='direct',recl=1,form='unformatted')   
   

  iw=0

do ig=1,9 !!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&!!!!!

    write(pnum,fmt='(I1)'),ig
    ifile = trim(ifile0)//'EF'//trim(pnum)
    call realinb(Is,Nz*Nx*NRHS/2,ifile)
   
    if (ig==1) then
      t1=MPI_Wtime()
      call find_scatters(Is, Nx, Nz, NRHS, nsi, xis, zis, si1, si2)
      allocate(Jig(nsi),Jall(nsi*ng))
      t2=MPI_Wtime()
      
      if (myrank==0) write(*,*) " "
      if (myrank==0) write(*,*) "Time for find_scatters ", t2-t1, "Error measures = ", nsi
    endif
    
  do deltax = 25,50,25  
  do winx = 10,100,30
  do sigmax = 5,45,10
  do winz = 100,100  !50,100,50
  do deltaz = 50,50  !!30,50,10
  do sigmaz = 15,15 !!5,45,10  
    !winz = 100 
    !deltaz = 50  
    !sigmaz = 15 
    
    Jig=0.
    t1=MPI_Wtime() 
      call power_norm2D_test(Jig,Is,Iadj,winx,winz,deltax,deltaz,sigmaz,sigmaz,Nx,Nz,NRHS,nsi,xis,zis,si1,si2)
    t2=MPI_Wtime()
    !if (myrank==0) write(*,*) "Time for power_norm is ", t2-t1
    
    Jsum=0.
    Jsum1=0.
    Jsum2=0.
    Jsum3=0.
    
    do ix=1,nsi
      Jsum = Jsum+Jig(ix)
      if (zis(ix) .lt. 250) then
        Jsum1 = Jsum1+Jig(ix)
      elseif (zis(ix) .lt. 375 .and. zis(ix) .ge. 250) then
        Jsum2 = Jsum2+Jig(ix)
      else     
        Jsum3 = Jsum3+Jig(ix)
      endif
    enddo

    iw = iw+1
    dt=t2-t1
 
    if (myrank==0) then
      write(unit=20,rec=iw)Jsum
      write(unit=21,rec=iw)Jsum1   
      write(unit=22,rec=iw)Jsum2
      write(unit=23,rec=iw)Jsum3
      write(unit=24,rec=iw)dt

      write(*,*) ig,winx,winz,deltax,deltaz,sigmax,sigmaz,Jsum,Jsum1,Jsum2,Jsum3,t2-t1
    endif
    
  enddo
  enddo
  enddo
  enddo
  enddo
  enddo




    deallocate(Jig,Jall)

enddo  !!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&!!!!!



  close(fps)
    
    
    
    
    

!! MPI Finalization
 call MPI_Finalize(ierr)



end program adjtest
