module context

use partition
implicit none
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  BLACS contexts construction module  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


    !! each process owns this information
    integer,allocatable :: ctxt(:), nprow(:), npcol(:), myrow(:), mycol(:), mynodes(:)


contains


    subroutine ctxt_generate( flag, lengbTT, gbTT, ctxt, nprow, npcol, myrow, mycol, mynodes )

        integer             :: flag, lengbTT
        type(TT_struct)     :: gbTT(lengbTT)
        integer,allocatable :: ctxt(:), nprow(:), npcol(:), myrow(:), mycol(:), mynodes(:)

        type :: usermap_struct
            integer,allocatable :: ary(:,:)
        end type usermap_struct
        type(usermap_struct),allocatable :: usermap(:)
        integer             :: Lvl, level, Ntmp, kidL, kidR, counter
        integer,allocatable :: row(:), col(:)


        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  generate context  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
        !! this mehod of generating sub-contexts are highly optimized  !!
        Lvl = gbTT(1)%LL
        allocate(ctxt(Lvl),nprow(Lvl),npcol(Lvl),myrow(Lvl),mycol(Lvl),mynodes(Lvl),usermap(lengbTT),row(lengbTT),col(lengbTT))
        
        !! generating mynodes
        Ntmp = flag
        do level = Lvl,1,-1
            mynodes(level) = Ntmp
            Ntmp = gbTT(Ntmp)%parent
        end do

        !! generating contexts by level
        row = 0
        col = 0
        counter = 0
        do Ntmp = 1,lengbTT

            level = gbTT(Ntmp)%LL

            if (gbTT(Ntmp)%kids(1) == 0) then
                row(Ntmp) = 1
                col(Ntmp) = 1
                allocate(usermap(Ntmp)%ary(row(Ntmp),col(Ntmp)))
                usermap(Ntmp)%ary(1,1) = counter    ! the process rank of the leaf nodes
                counter = counter + 1
            else
                kidL = gbTT(Ntmp)%kids(1)
                kidR = gbTT(Ntmp)%kids(2)
                if (mod(Lvl-gbTT(Ntmp)%LL,2) == 0) then    ! square mesh grid
                    row(Ntmp) = row(kidL) * 2
                    col(Ntmp) = col(kidL)
                    allocate(usermap(Ntmp)%ary(row(Ntmp),col(Ntmp)))
                    usermap(Ntmp)%ary(           1:row(kidL), 1:col(Ntmp) ) = usermap(kidL)%ary
                    usermap(Ntmp)%ary( row(kidL)+1:row(Ntmp), 1:col(Ntmp) ) = usermap(kidR)%ary
                else    ! rectangular mesh grid
                    row(Ntmp) = row(kidL)
                    col(Ntmp) = col(kidL) * 2
                    allocate(usermap(Ntmp)%ary(row(Ntmp),col(Ntmp)))
                    usermap(Ntmp)%ary( 1:row(Ntmp),           1:col(kidL) ) = usermap(kidL)%ary
                    usermap(Ntmp)%ary( 1:row(Ntmp), col(kidL)+1:col(Ntmp) ) = usermap(kidR)%ary
                end if
                deallocate( usermap(kidL)%ary, usermap(kidR)%ary )
            end if           
 
            ! BLACS contexts generating 
            if (mynodes(level) == Ntmp) then
                call blacs_get( 0, 0, ctxt(level) )
                call blacs_gridmap( ctxt(level), usermap(Ntmp)%ary, row(Ntmp), row(Ntmp), col(Ntmp) )
                call blacs_gridinfo( ctxt(level), nprow(level), npcol(level), myrow(level), mycol(level) )
            end if

        end do
        deallocate(usermap(lengbTT)%ary)
        deallocate(usermap,row,col)


        return
    end subroutine ctxt_generate


end module context
