program adjtest

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!  2D massively parallel HSS multifontal Helmholtz solver: HelmPar2D  !!!!!!!!!!!!!!!
!!!!!!!!!!!           Shen Wang,  Jianlin Xia,  Maarten V. de Hoop              !!!!!!!!!!!!!!!
!!!!!!!!!!!                Purdue & ConocoPhillips, June 2010                   !!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

use NDpre2D
use context
use A2D 
use fact
use solution2D
use parser
use fileio
use powernorm
use inversion

implicit none
include 'mpif.h'


!! declaring variables
integer                      ::  ierr, myrank, nprocs
integer                      ::  JOB, tos, passive
integer                      ::  Lvl, slvl, Lpar, rkmin, mr2d,start,FS
real(kind=4)                 ::  tol, dt, dw, step,sumbase
real(kind=8)                 ::  t1, t2
double precision             :: Jsum,Jsum0
integer                      ::  Nz, Nx,nvar, N_delta, NRHS, Nt, Nw,Nh,lnx,lnz,Np,nsi,Ng, fps=20, iter,niter,nsteps,iflag,istep,igg
 complex(kind=4)             :: sigma0, hz, hx, tmpi,tmpdi,tmpdx,tmpdz,tmpg,tmpgpx,tmpgmx,tmpgpz,tmpgmz,dPhi,tmpss
integer,allocatable          ::  sourceZ(:), sourceX(:),h(:),arrayloc(:),sources(:)
real(kind=4),allocatable     ::  vel(:),vel0(:), dv(:), ga(:), gz(:),gx(:), taper(:), N(:),water(:),waterpr(:),rayp(:)
real,allocatable ::  vupdate(:),grad(:),grad0(:),vtmp(:),lbfgs_work(:),lbfgs_diag(:),rad(:),tht(:)
real(kind=4),allocatable     ::  Jig(:),Jall(:),steps(:)
 complex(kind=4),allocatable  ::  Y(:),Y0(:),Y1(:),YY(:), freq(:),Aw(:),dataw(:),dataws(:),fieldout(:),fieldout0(:)
 complex(kind=4),allocatable  ::  X(:,:,:),X0(:,:,:) 
real(kind=4)                 ::  tmp, tmpz, tmpx,alpha,maxp,xis(1000000),zis(1000000),si1(1000000),si2(1000000),betaFR,base(99,99),zp(99),xp(99)
integer                      ::  rhs, zz, xx, pp, iw, w0, wf,xi,zi, ig,ds,dh,ihh,hh, iss, ix, iz, headZ, headX,forward,adjyn
 character(len=256)           ::  vfile,v0file,dvfile,sfile,sfile0,rfile,rfile0,ifile,ifile0,wnum,pnum 
 !!complex(kind=4),parameter   :: ONE = (1.0E+0,0.0E+0), ZERO = (0.0E+0,0.0E+0), RT = (0.0E+0,1.0E+0)
real(kind=4),allocatable     ::  vel_pz(:,:), dvel_pz(:,:), rho(:,:),rad_pr(:),tht_pr(:)

  
  
!! MPI Initialization
 call MPI_Init(ierr)
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
 call MPI_Comm_size( MPI_COMM_WORLD, nprocs, ierr )

if (myrank==0) write(*,*) "nprocs is ",nprocs

!! set up variables
 call frompar('level',Lvl,14)
 call frompar('Nz',Nz,2048)
 call frompar('Nx',Nx,2048)
 !call frompar('Ns',NRHS,1)
 call frompar('Nt',Nt,2048)
 call frompar('Nw',Nw,55)
 call frompar('Nh',Nh,20)
 call frompar('dh',dh,2)
 call frompar('dx',tmp,250.); hx=tmp
 call frompar('dz',tmp,250.); hz=tmp
 call frompar('dt',dt,0.01)
 call frompar('ds',ds,10)
 call frompar('w0',w0,2)
 call frompar('wf',wf,Nw)
 call frompar('x0',xi,80)
 call frompar('z0',zi,110)
 call frompar('alpha',alpha,1.)
 call frompar('vfile',vfile)
 call frompar('v0file',v0file)
 call frompar('dvfile',dvfile)
 call frompar('sfile',sfile0)
 call frompar('rfile',rfile0,' ')
 call frompar('ifile',ifile0,' ')
 call frompar('passive',passive,0)
 !call frompar('Ns',NRHS,29)
 call frompar('maxp',maxp,30.)
 call frompar('forward',forward,1)
 call frompar('adjoint',adjyn,1) 
 call frompar('niter',niter,1) 
 call frompar('startiter',start,1)
 call frompar('nscat',nsi,1)
 call frompar('slownesses',Np,15)
 call frompar('freesurf',FS,0)
 
!! choosing jobs. 1: isotropic acoustic, 2: elliptic anisotropy, 3: VTI, 4. HTI, 5. TTI
JOB = 1
tos = 1

NRHS=Np*2-1



  allocate(vel(Nz*Nx),vel0(Nz*Nx),dv(Nz*Nx))
  allocate(Jall(niter))


  vel = 0.
  vel0 = 0.
  dv = 0.
  Jall = 0.

 
allocate(Y(Nz*Nx*NRHS))
allocate(Y0(Nz*Nx*NRHS),Y1(Nx*Nz*NRHS))
!allocate(X(Nz,Nx,NRHS),X0(Nz,Nx,NRHS))
allocate(dataw(Nx*NRHS))
allocate(dataws(Nx*NRHS))
allocate(freq(Nw),Aw(Nw))
allocate(sourceZ(NRHS),sourceX(NRHS),rayp(NRHS))         
allocate(taper(Nx),arrayloc(Nx))
allocate(ga(ng), gx(ng), gz(ng))
allocate(sources(NRHS))
 
 
allocate(rad(Nz), tht(Nx))
hz             = (250.0E+0,0.0E+0)
hx	       =  pi/8/Nx
do zz=1,Nz
  rad(zz)=1000000 - real(hz)*(zz-1)
enddo
do xx=1,Nx
  tht(xx) = 7.*pi/16. + hx*(xx-1)
enddo
if (myrank==0) write(*,*) "Debug 1",hx,tht(1),tht(Nx),rad(1),rad(Nz)

   
    
!! read in velocity model

  call realinb(vel0,Nx*Nz,v0file)   ! Smooth starting model
  call realinb(vel,Nx*Nz,vfile)      ! True perturbation
  
  vel0 = vel0 + vel                  ! vel0 is now true smooth model
  
  call realinb(dv,Nx*Nz,dvfile)     ! Perturbation for test
  
 
         
  if (myrank==0) write(*,*) "Loaded velocity models"
 
     
sourceZ = 0
sourceX = 0
do rhs=1,NRHS
    sourceZ(rhs) = zi 
    sourceX(rhs) = xi + ds*(rhs-1)
    if (passive == 1) rayp(rhs) = sin(maxp*pi/180)/vel0(zi+nz*(xi-1))*(rhs-Np)
    if (myrank==0) write(*,*) rayp(rhs)
end do  
              
  dw = 2*PI/(Nt*dt)
do iw=1,Nw
  freq(iw)=(iw-1)*dw
  Aw(iw) = exp(-(real(wf+w0)/2 - real(iw))**2/(real(wf-w0)/5)**2)
enddo
sigma0 = 1.0*freq(wf)

  
arrayloc = 0
!do xx=100,Nx-100,20
do xx=100,Nx-100,1
  arrayloc(xx)=1
enddo

taper=0.
do zz=1,Nx
  if (arrayloc(zz)==1) then
    do xx=1,Nx
      taper(xx)=taper(xx)+exp(-real(xx-zz)**2/2**2.)
    enddo
  endif
enddo
taper = taper/maxval(taper,Nx)     
    
 

 deallocate(arrayloc)


!! set up global variables
!Lvl            = 16
slvl           = 0
rkmin          = 20                                       ! minimum rank to guarantee the numerical stability
tol            = 1.0E-4
Lpar           = nint(log(real(nprocs))/log(2.0)) + 1     ! Lpar = log_2(nprocs) + 1
mr2d           = 0
N_delta        = 50




  
  if (myrank==0) write(*,*) " "
  if (myrank==0) write(*,*) " "  
  if (myrank==0 .and. adjyn==1) write(*,*) "IMAGING FOR PERTURBATION ",ig
  

if (myrank==0) then
write(*,*) "Data source: ",trim(sfile0),myrank
write(*,*) "Image file:  ",trim(ifile0)
endif


if (myrank==0) write(*,*) "Number of right hand sides", NRHS



Y=0.

dataw=0.




  



do iw=w0,wf

!! Define initial conditions

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  1. 2D nested dissection preprocessing  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
 call NDpre_generate( myrank, tos, Lvl, Lpar, Nz, Nx, lengbTT, lenlocTT, gbTT, locTT, gbTlabel, locTlabel, gbNB, locNB, locHZ, locTZ, locHX, locTX, label, flag )
t2 = MPI_Wtime()
!if (myrank == 0) then
    !write(*,*) "2D nested dissection domain decomposition is done! wall time is ", t2-t1
!end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2. BLACS contexts generating  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 call ctxt_generate( flag, lengbTT, gbTT, ctxt, nprow, npcol, myrow, mycol, mynodes ) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!! find vel model for processor
    allocate(vel_pz(locTZ-locHZ+1,locTX-locHX+1))
    allocate(dvel_pz(locTZ-locHZ+1,locTX-locHX+1))

  
  allocate(rho(locTZ-locHZ+1+2*tos,locTX-locHX+1+2*tos))
  rho = 1.0E+0



  allocate(tht_pr(locTX-locHX+1+2*tos),rad_pr(locTZ-locHZ+1+2*tos))
  do xx = locHX-tos,locTX+tos
    tht_pr(xx-locHX+1+tos) = tht(xx)
  enddo
 
  do zz = locHZ-tos, locTZ+tos
    rad_pr(zz-locHZ+1+tos) = rad(zz)
  enddo
 





Y = ZERO



!! find vel model for processor
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    vel_pz(zz-locHZ+1,xx-locHX+1) = vel0(zz+Nz*(xx-1)) + dv(zz+Nz*(xx-1)) !true model
        end do
    end do



if (myrank == 0) write(*,*) "Beginning to propagate"


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. generating the matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
if (JOB == 1) then        ! isotropic acoustic
    call A2D_acoustic_sphere( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho,rad_pr,tht_pr, A)
   ! call A2D_acoustic( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, A )
endif
t2 = MPI_Wtime()

if (myrank == 0) then
    write(*,*) "local matrix A generation is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!t1 = MPI_Wtime()
 call factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                   lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
   write(*,*) "The total factorization is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4A. define initial wavefields !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


Y = ZERO

taper=0.
do zz=101,Nx-100
  do xx=1,Nx
    taper(xx)=taper(xx)+exp(-real(xx-zz)**2/50**2.)
  enddo
enddo
taper = taper/maxval(taper,Nx)
do rhs = 1,NRHS
    do xx = 1,Nx
        do zz = 1,Nz
            tmp = real((zz-sourceZ(rhs))**2)
	    !tmp = real((zz-sourceZ(rhs))**2+(xx-sourceX(rhs))**2)
            if (tmp <= 25.0E+0) then
            Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = taper(xx)*exp(RT*hx*850000*xx*rayp(rhs)*real(freq(iw))*dt*10)*exp(-tmp/4.0)
            end if
        end do
    end do
end do


	 
 
 if (myrank == 0) then
  write(wnum,fmt='(I2)')iw
  sfile=trim(sfile0)//'s'//trim(wnum)
  
  dataw = ZERO
  do xx=1,Nx
   do rhs=1,NRHS
      dataw(xx + Nx*(rhs-1)) = Y(sourceZ(rhs) + Nz*(xx-1) + Nz*Nx*(rhs-1))
   enddo
  enddo
  
  call complexoutb(dataw,Nx*NRHS,sfile)
endif

if (myrank==0) write(*,*) "Made initial wavefields"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5. solution  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
lnx=locTX-locHX+1
lnz=locTZ-locHZ+1
allocate( X(lnz,lnx,NRHS))


! write(*,*) "debuga"
X = ZERO
do rhs = 1,NRHS
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    X(zz-locHz+1,xx-locHx+1,rhs) = Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1))  	    
        end do
    end do
end do
if (locHZ > 1) then   ! CAUTION!!
    X(1:tos,:,:) = ZERO
end if
if (locHX > 1) then
    X(:,1:tos,:) = ZERO
end if
 !write(*,*) "debugb"
 
 
t1 = MPI_Wtime()
 call mfsolve(X,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
t2 = MPI_Wtime()
if (myrank == 0) then
    write(*,*) "The wall time of the solution to", NRHS, " RHS is ", t2-t1
        write(*,*) "Debug 2 ",X(Nz/2,Nx/2,16)
end if

 call factors_dealloc( lenlocTT, locUL, locUR, locLL, locIPIV, Lpar-1, gbUL, gbUR, gbLL, gbIPIV )
 deallocate(locUL, locUR, locLL, locIPIV, gbUL, gbUR, gbLL, gbIPIV )  

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  6. Imaging  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 
   headZ = locHZ
   if (headZ > 1) then
        headZ = headZ + tos
   end if
   headX = locHX
   if (headX > 1) then
       headX = headX + tos
   end if

  allocate(YY(Nx*Nz*NRHS))
  YY=0.
  
do rhs = 1,NRHS
    do xx = headX,locTX
        do zz = headZ,locTZ
	    YY(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = X(zz-locHz+1,xx-locHx+1,rhs)
        end do
    end do
end do

 Y1=0.

 call MPI_ALLREDUCE(YY,Y1,Nx*Nz*NRHS,MPI_COMPLEX,MPI_SUM,MPI_COMM_WORLD,ierr)


 !deallocate(gbTT,locTT,gbTlabel,locTlabel,label,locNB,gbNB)
 !deallocate(ctxt, nprow, npcol, myrow, mycol, mynodes)
 
 deallocate(YY)
















if (myrank==0) then
!write(*,*) " "
write(*,*) "Propagating for frequency ",iw
!write(*,*) "********************************"
endif

    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    vel_pz(zz-locHZ+1,xx-locHX+1) = vel0(zz+Nz*(xx-1)) !smooth model
        end do
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. generating the matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
if (JOB == 1) then        ! isotropic acoustic
    call A2D_acoustic_sphere( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho,rad_pr,tht_pr, A)
    !call A2D_acoustic( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, A )
endif
t2 = MPI_Wtime()

if (myrank == 0) then
    write(*,*) "local matrix A generation is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
 call factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                   lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
t2 = MPI_Wtime()
if (myrank == 0) then
   write(*,*) "The total factorization is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4A. define initial wavefields !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


Y = ZERO
!Y0 = ZERO


taper=0.
do zz=101,Nx-100
  do xx=1,Nx
    taper(xx)=taper(xx)+exp(-real(xx-zz)**2/50**2.)
  enddo
enddo
taper = taper/maxval(taper,Nx)
do rhs = 1,NRHS
    do xx = 1,Nx
        do zz = 1,Nz
            tmp = real((zz-sourceZ(rhs))**2)
	    !tmp = real((zz-sourceZ(rhs))**2+(xx-sourceX(rhs))**2)
            if (tmp <= 25.0E+0) then
            Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = taper(xx)*exp(RT*hx*850000*xx*rayp(rhs)*real(freq(iw))*dt*10)*exp(-tmp/4.0)
            end if
        end do
    end do
end do

write(*,*) "Loaded initial wavefields"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5. solution  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
lnx=locTX-locHX+1
lnz=locTZ-locHZ+1
allocate( X0(lnz,lnx,NRHS))


! write(*,*) "debuga"
X0 = ZERO
do rhs = 1,NRHS
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    X0(zz-locHz+1,xx-locHx+1,rhs) = Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1))  !!Y(zz,xx,rhs)	    
        end do
    end do
end do
if (locHZ > 1) then   ! CAUTION!!
    X0(1:tos,:,:) = ZERO
end if
if (locHX > 1) then
    X0(:,1:tos,:) = ZERO
end if
 !if(myrank==0) write(*,*) "debugb"
t1 = MPI_Wtime()
 call mfsolve(X0,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
t2 = MPI_Wtime()
if (myrank == 0) then
    write(*,*) "The wall time of the solution to", NRHS, " RHS is ", t2-t1
    write(*,*) "Debug 2 ",X0(Nz/2,Nx/2,16)
end if




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  6. Imaging  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 
   headZ = locHZ
   if (headZ > 1) then
        headZ = headZ + tos
   end if
   headX = locHX
   if (headX > 1) then
       headX = headX + tos
   end if

  allocate(YY(Nx*Nz*NRHS))
  YY=0.
  
do rhs = 1,NRHS
    do xx = headX,locTX
        do zz = headZ,locTZ
	    YY(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = X0(zz-locHz+1,xx-locHx+1,rhs)
        end do
    end do
end do

 Y0=0.

 !write(*,*) "Debug3 ",myrank, YY(Nz/2+Nz*(Nx/2-1)+Nz*Nx*(16-1))

 call MPI_ALLREDUCE(YY,Y0,Nx*Nz*NRHS,MPI_COMPLEX,MPI_SUM,MPI_COMM_WORLD,ierr)
 
 !if (myrank==0) write(*,*) "Debug3AAAAA ",myrank, Y0(Nz/2+Nz*(Nx/2-1)+Nz*Nx*(16-1))

 deallocate(gbTT,locTT,gbTlabel,locTlabel,label,locNB,gbNB)
 deallocate(ctxt, nprow, npcol, myrow, mycol, mynodes)
 
 deallocate(YY,vel_pz,dvel_pz,rho,rad_pr,tht_pr)








 if (myrank == 0) then
  
  sfile=trim(sfile0)//trim(wnum)
  dataw = ZERO
  do xx=1,Nx
   do rhs=1,NRHS
      dataw(xx + Nx*(rhs-1)) = Y1(sourceZ(rhs) + Nz*(xx-1) + Nz*Nx*(rhs-1)) - Y0(sourceZ(rhs) + Nz*(xx-1) + Nz*Nx*(rhs-1)) 
   enddo
  enddo
  
  call complexoutb(dataw,Nx*NRHS,sfile)
  
 endif
 if (myrank==0) write(*,*) "Debug4 ", dataw((Nx/2)+Nx*(16-1))
 
 call factors_dealloc( lenlocTT, locUL, locUR, locLL, locIPIV, Lpar-1, gbUL, gbUR, gbLL, gbIPIV )
 deallocate(X,X0,locUL, locUR, locLL, locIPIV, gbUL, gbUR, gbLL, gbIPIV )  
 
  t2 = MPI_Wtime()
  if (myrank == 0 ) write(*,*) "Frequency",iw," Wall time for modeling ", t2-t1

enddo !!!%%%%%%%%%%%%%%%%%%%%%%%END LOOP OVER FREQUENCY%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!!
 



  
  
  
    deallocate(Y,Y0,Y1)
    deallocate(ga)   
    deallocate(freq)
    deallocate(Aw)
    deallocate(sourceX)
    deallocate(sourceZ)
    deallocate(dataw,dataws) 
    deallocate(taper)
    deallocate(rayp)
    deallocate(gz )  
    deallocate(sources)
    deallocate(rad,tht) 	
    deallocate(vel,Jall,vel0,dv)

 call MPI_Barrier(MPI_COMM_WORLD,ierr)



!! MPI Finalization
 call MPI_Finalize(ierr)



end program adjtest
