



MODULE parser

integer,private,parameter      :: cdim=4096
integer,private,parameter      :: charlen=256
integer,private                    :: iparm=0,iunit=555
character(100),private       :: parsed_param(1000)
character(1) :: blank=" ",guillemet="""",tab="	"
logical, private     :: debug=.false.

interface frompar
  module procedure frompar_string
  module procedure frompar_integer
  module procedure frompar_logical
  module procedure frompar_real
end interface


contains

subroutine frompar_string(keyword,value,default)
  implicit none
  character(*),intent(in)           :: keyword
  character(*),intent(out)          :: value
  character(*),intent(in),optional  :: default
  logical          :: lerr, lerrg, lerrb
  character(cdim)  :: valueb, valueg
  integer          :: ilb, ilg, iposg, iposb, len_value, ierr
 
  value=""
  call fromparstr(keyword,valueg,lerrg,delimiter=guillemet,il=ilg, ipos=iposg)
  call fromparstr(keyword,valueb,lerrb,delimiter=blank,il=ilb, ipos=iposb)
  if (TRIM(valueb)==(guillemet//guillemet)) lerrb=.true.
  lerr=(lerrg.and.lerrb)
  if (lerr) then
    if (present(default)) then
      value=TRIM(default)
    else
      print*,'Error from FromPar.  Cannot read value associated with keyword ',TRIM(keyword)
      call exit(-1)
    endif
  else if (lerrg) then
    value=valueb
  else if (lerrb) then
    value=valueg
  else
    if (ilg>ilb) then
      value=valueg
    else if (ilg==ilb) then
      if (iposg>iposb) then
        value=valueg
      else
        value=valueb
      end if
    else
      value=valueb
    endif
  endif
  len_value=len(value)

  value=trim(value)//" "
  iparm=iparm+1
  parsed_param(iparm)="par: "//TRIM(keyword)//"="//TRIM(value)

end subroutine frompar_string



subroutine frompar_integer(keyword,value,default)

  implicit none
  character(*),intent(in)           :: keyword
  integer,intent(out)          :: value
  integer,intent(in),optional  :: default
  logical          :: lerr, lerr2
  character(charlen)  :: svalue,newsvalue
  integer          :: ierr,length

  call fromparstr(keyword,svalue,lerr)
  if (lerr) then
    if (present(default)) then
      value=default
    else
       print*,'Error from FromPar.  Cannot read value associated with keyword ',TRIM(keyword)
      call exit(-1)
    endif
  else
!    print*,'SVALUE IS ',svalue
    value=string2integer(svalue,lerr2)
    if (lerr2) then
      print*,'Error from FromPar.  Bad value associated with keyword ',TRIM(keyword)
      call exit(-1)
    endif
  endif

  write(svalue,*)value
  iparm=iparm+1
  parsed_param(iparm)="par: "//TRIM(keyword)//"="//TRIM(ADJUSTL(svalue))

end subroutine frompar_integer



subroutine frompar_logical(keyword,value,default)

  implicit none
  character(*),intent(in)           :: keyword
  logical,intent(out)          :: value
  integer,intent(in),optional  :: default
  logical          :: lerr, lerr2
  character(charlen)  :: svalue
  integer          :: ierr

  call fromparstr(keyword,svalue,lerr)
  if (lerr) then
    if (present(default)) then
      value=(default==1)
    else
       print*,'Error from FromPar.  Cannot read value associated with keyword ',TRIM(keyword)
      call exit(-1)
    endif
  else
    value=(string2integer(svalue,lerr2)==1)
    if (lerr2) then
      print*,'Error from FromPar.  Bad value associated with keyword ',TRIM(keyword)
      call exit(-1)
    else if (TRIM(svalue)/="0" .and. TRIM(svalue)/="1") then
      print*,'Error from frompar.  Value read for keyword ',TRIM(keyword),' different from 0 or 1'
      call exit(-1)
    end if
  end if

  write(svalue,*)value
  iparm=iparm+1
  parsed_param(iparm)="par: "//TRIM(keyword)//"="//TRIM(ADJUSTL(svalue))

end subroutine frompar_logical



subroutine frompar_real(keyword,value,default)

  implicit none
  character(*),intent(in)           :: keyword
  real,intent(out)          :: value
  real,intent(in),optional  :: default
  logical          :: lerr, lerr2
  character(charlen)  :: svalue
  integer          :: ierr

  call fromparstr(keyword,svalue,lerr)
  if (lerr) then
    if (present(default)) then
      value=default
    else
       print*,'Error from FromPar.  Cannot read value associated with keyword ',TRIM(keyword)
      call exit(-1)
    endif
  else
    value=string2real(svalue,lerr2)
    if (lerr2) then
      print*,'Error from FromPar.  Bad value associated with keyword ',TRIM(keyword)
      call exit(-1)
    endif
  endif

  write(svalue,*)value
  iparm=iparm+1
  parsed_param(iparm)="par: "//TRIM(keyword)//"="//TRIM(ADJUSTL(svalue))

end subroutine frompar_real






subroutine fromparstr(keyword,svalue,lerr,delimiter,il,ipos)
   implicit none
   
  character(*),intent(in)           :: keyword
  character(*),intent(out)          :: svalue
  logical,intent(out)               :: lerr
  character(*),intent(in),optional          :: delimiter
  integer,intent(out),optional          :: il,ipos

  character(1)     :: delim
  character(cdim)  :: commandline, parfile, svalue1,svalue2
  logical          :: lerrpar, lerr1, lerr2
  integer          :: ils, ils1, iposs, iposs1,iposs2


  delim=blank
  if (present(delimiter)) delim=delimiter

  lerr=.false.
  call getcommandline(commandline)
  iposs = string2par(commandline,"par",parfile,lerrpar,blank)
  lerr1=.true.

  if (.not.lerrpar) &
    call fromfilestring(parfile,keyword,svalue1,lerr1,delim,ils1,iposs1)

  iposs2 = string2par(commandline,keyword,svalue2,lerr2,delim)
  if (.not.lerr1) then
    svalue=svalue1
    iposs=iposs1
    ils=ils1
  endif
!print*,'SVALUE1 = ',TRIM(svalue1)
  if (.not.lerr2) then
    svalue=svalue2
    iposs=iposs2
    ils=9999999
  endif
!print*,'SVALUE2 = ',TRIM(svalue2)

  if (lerr1.and.lerr2) lerr=.true.
  if (.not.lerr) then
     if (present(il)) il=ils
     if (present(ipos)) ipos=iposs
  else
     if (present(il)) il=0
     if (present(ipos)) ipos=0
  endif


end subroutine fromparstr


subroutine fromfilestring(filename,keyword,svalue,lerr,delimiter,il,ipos)
  implicit none
  character(*), intent(in)          :: filename,keyword
  character(*),intent(out)         :: svalue
  logical, intent(out)               :: lerr
  character(1),intent(in),optional   :: delimiter
  integer, intent(out),optional      :: il,ipos

  character(cdim)   :: temp,svaluetemp,temp2
  logical             :: lerr2
  character(1)   :: delim
  integer             :: iposs,ils,iposs2

  delim=blank
  if (present(delimiter)) delim=delimiter

  lerr=.true.
  open(unit=iunit,file=TRIM(filename),form='formatted',err=90)

  ils=1
  do 
    read(iunit,100,end=80) temp
    iposs2=index(TRIM(temp),"#",.false.)
    if (iposs2==0) iposs2=len_trim(temp)

    iposs=string2par(temp(1:iposs2),TRIM(keyword),svaluetemp,lerr2,delim)
    if (.not.lerr2) then
      svalue=svaluetemp
      lerr=.false.
      if (present(il)) il=ils
      if (present(ipos)) ipos=iposs
    endif
    ils=ils+1
  enddo

  if (lerr) then
     if (present(il)) il=0
     if (present(ipos)) ipos=0
  endif

  90 write(0,*) 'Error opening file ',TRIM(filename)
        call exit(-1)
  80 close(iunit)
  100 format(a2048)

end subroutine fromfilestring



integer function string2par(string,keyword,svalue,lerr,delimiter)
  implicit none
  character(*), intent(in)          :: string,keyword
  character(*),intent(out)         :: svalue
  logical, intent(out)               :: lerr
  character(1),intent(in),optional   :: delimiter

  logical          :: lerr1,lerr2,lerr3,lerr4
  integer          :: ipos1,ipos2,ipos3,ipos4, ipos
  character(cdim)    :: svalue1,svalue2,svalue3,svalue4

  ipos=0; lerr=.true.

  if (present(delimiter)) then
    ipos1=string2par2(string,keyword,svalue1,lerr1,blank,tab,delimiter)
    ipos2=string2par2(string,keyword,svalue2,lerr2,blank,blank,delimiter)
    ipos3=string2par2(string,keyword,svalue3,lerr3,tab,tab,delimiter)
    ipos4=string2par2(string,keyword,svalue4,lerr4,tab,blank,delimiter)
  else
    ipos1=string2par2(string,keyword,svalue1,lerr1,blank,tab)
    ipos2=string2par2(string,keyword,svalue2,lerr2,blank,blank)
    ipos3=string2par2(string,keyword,svalue3,lerr3,tab,tab)
    ipos4=string2par2(string,keyword,svalue4,lerr4,tab,blank)
  endif
!print*,'different svalues are: ',trim(svalue1),trim(svalue2),trim(svalue3),trim(svalue4)

  if (.not.lerr1) then
    lerr=.false.
    ipos=max(ipos,ipos1)
    svalue=TRIM(svalue1)
  endif
  if (.not.lerr2) then
    lerr=.false.
    ipos=max(ipos,ipos2)
    svalue=TRIM(svalue2)
  endif
  if (.not.lerr3) then
    lerr=.false.
    ipos=max(ipos,ipos3)
    svalue=TRIM(svalue3)
  endif
  if (.not.lerr4) then
    lerr=.false.
    ipos=max(ipos,ipos4)
    svalue=TRIM(svalue4)
  endif

  string2par=ipos
end function string2par



integer function string2par2(string,keyword,svalue,lerr,lsep,rsep,delimiter)
  implicit none
  character(*), intent(in)          :: string,keyword
  character(*),intent(out)         :: svalue
  logical, intent(out)               :: lerr
  character(1),intent(in)        :: lsep, rsep
  character(1),intent(in),optional   :: delimiter

  character(cdim+2)     :: temp,temp2,stringtemp
  integer                         :: ipos,iend,lkey
  character(1)                :: deliml,delimr

  if (present(delimiter)) then
    deliml=delimiter
    delimr=delimiter
  else
    deliml=""
    delimr=rsep
  endif

  stringtemp=lsep//TRIM(ADJUSTL(string))//rsep
  temp=lsep//TRIM(keyword)//"="//deliml
  lerr=.false.
  lkey=len_trim(keyword)+1+1+len_trim(deliml)

  if (debug) print*,"string2par2: look for",trim(temp),' in: ',trim(stringtemp)

  ipos=index(TRIM(stringtemp),TRIM(temp),.true.)
  iend=index(stringtemp(ipos+lkey:),delimr,.false.)

  if (ipos==0 .or. iend==0 .or. (ipos+iend-1)>cdim) then
    lerr=.true.
    svalue=""
    string2par2=0
    return
  endif
  iend=iend+ipos+lkey-2
  ipos=ipos+lkey
  if (ipos>iend) then
    lerr=.true.
    svalue=""
    string2par2=0
  else
    svalue=stringtemp(ipos:iend)
    string2par2=ipos
  endif

end function string2par2




subroutine getcommandline(string)
  implicit none
  character(*),intent(out)   :: string
  integer                 :: i, iargc
  character(cdim)  :: temp
  string=""
  do i=1,iargc()
    call getarg(i,temp)
    string=TRIM(string)//" "//TRIM(temp)
  enddo
end subroutine getcommandline


real function string2real(string,lerr)
  implicit none
  character(*), intent(in)      :: string
  logical,intent(out),optional  :: lerr
  real   :: rvalue
  if (present(lerr)) lerr=.false.
  read(string,*,err=80) rvalue
  string2real=rvalue
  return
80 if (present(lerr)) lerr=.true.
end function string2real

integer function string2integer(string,lerr)
  implicit none
  character(*), intent(in)      :: string
  logical,intent(out),optional  :: lerr
  if (present(lerr)) lerr=.false.
!  print*,"the string is ",string
  read(string,*,err=80) string2integer
  return
80 if (present(lerr)) lerr=.true.
end function string2integer






end module parser
