module powernorm
contains

!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!

subroutine power_norm(J,I,adjI,Nx,Nz,NRHS,nsi,xi,zi,si1,si2)


  use parser
  use fileio 

  implicit none
  
  include 'mpif.h'
  
  integer        :: ix,iz,nn,ixx,dz, is1,is2,myrank,ierr,nprocs
  integer        :: nx, nz, nrhs, nsi, winz, W
  real           :: I(Nx*Nz*NRHS),adjI(Nx*Nz*NRHS/2), sigma, S
  real           :: J(nsi),xi(*),zi(*),si1(*),si2(*)
  real, allocatable   :: corr(:),w2(:),mu1(:)
  !character(256) :: Ifile,srcnum
  
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
  
  call frompar('winz',winz,100)
  call frompar('deltaz',W,50)
  call frompar('sigma',sigma,40.)

  allocate(corr(2*W+1),w2(2*W+1),mu1(2*W+1))

 
 
 J = 0.
 do nn=1,nsi
   ix=Nz*(xi(nn)-1)
   is1=Nz*Nx*(si1(nn)-1)
   is2=Nz*Nx*(si2(nn)-1)
   
	corr=0.
	S=0.
	do dz=-W,W
	  do iz=zi(nn)-winz/2,zi(nn)+winz/2
            if (iz+dz .lt. Nz) corr(dz+W+1)=corr(dz+W+1)+I(iz+dz+ix+is1)*I(iz+ix+is2)
          enddo
          !!w1(dz+W+1) = (dz/W)**2;
          w2(dz+W+1) = -exp(-(dz**2)/sigma**2)
	  S=S+corr(dz+W+1)**2
   	enddo
	
	do dz=-W,W
    	  J(nn) = J(nn) + S**(-1)*w2(dz+W+1)*(corr(dz+W+1)**2);
	enddo

	mu1=0.
	do dz=-W,W
    	  mu1(dz+W+1) = mu1(dz+W+1) - 2*S**(-1)*(w2(dz+W+1)-J(nn))*corr(dz+W+1);
	enddo

	do dz=-W,W
	  do iz=zi(nn)-winz/2,zi(nn)+winz/2
            if (iz+dz .lt. Nz) then
	      !adjI(iz+dz+ix+is1) = adjI(iz+dz+ix+is1) + mu1(dz+W+1)*I(iz+ix+is2)
	      adjI(iz+ix+is2)    = adjI(iz+ix+is2)    + mu1(dz+W+1)*I(iz+dz+ix+is1)
	    endif
	  enddo
	enddo
	
	
  enddo



  deallocate(corr, w2, mu1)





end subroutine power_norm


!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!


subroutine power_norm_adjtest(J,I,adjI,Nx,Nz,NRHS,nsi,xi,zi,si1,si2)


  use parser
  use fileio 

  implicit none
  
  include 'mpif.h'
  
  integer        :: ix,iz,nn,ixx,dz,inn, is1,is2,myrank,ierr,nprocs
  integer        :: nx, nz, nrhs, nsi, winz, W, ntap
  real           :: I(Nx*Nz*NRHS),adjI(Nx*Nz*NRHS/2), dI(Nx*Nz*NRHS/2), sigma, S,S0,hi1(Nz),hi2(Nz)
  real           :: totdidi, Jsum
  real           :: J(nsi),J0(nsi),didi(nsi),xi(*),zi(*),si1(*),si2(*)
  real, allocatable   :: corr(:),dcorr(:), w2(:),mu1(:),taper(:)
  !character(256) :: Ifile,srcnum
  real(kind=4),parameter   :: pi = 3.1415926E+0
  
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
  
  call frompar('winz',winz,100)
  call frompar('deltaz',W,50)
  call frompar('sigma',sigma,40.)

  allocate(corr(2*W+1),dcorr(2*W+1),w2(2*W+1),mu1(2*W+1),taper(winz))

  ntap = 20
  taper=1.
  do ix=1,ntap
    taper(ix) = cos((ix-1)/(ntap-1)*pi/2)**2
    taper(ntap+1-ix) = cos((ix-1)/(ntap-1)*pi/2)**2
  enddo
 
 do ix=1,Nx
   do iz = 1,Nz
     do nn=1,NRHS/2
          ixx=Nz*(ix-1)
	  inn=Nx*Nz*(nn-1)
	  dI(iz+ixx+inn) = I(iz+ixx+inn)-I(iz+ixx+inn+Nx*Nz*NRHS/2)
     enddo
   enddo
 enddo
    
 
 
 
 J = 0.
 J0 = 0.
 Jsum = 0.
 didi = 0.
 totdidi=0.
 do nn=1,nsi
   ix=Nz*(xi(nn)-1)
   is1=Nz*Nx*(si1(nn)-1)
   is2=Nz*Nx*(si2(nn)-1)
   
	corr=0.
	dcorr=0.
	S=0.
	S0=0.
	do dz=-W,W
	  do iz=zi(nn)-winz/2,zi(nn)+winz/2
            if (iz+dz .lt. Nz) corr(dz+W+1)=corr(dz+W+1)+I(iz+dz+ix+is1)*I(iz+ix+is2)
	    if (iz+dz .lt. Nz) dcorr(dz+W+1)=dcorr(dz+W+1)+I(iz+dz+ix+is1)*dI(iz+ix+is2)
	    if (iz+dz .lt. Nz) dcorr(dz+W+1)=dcorr(dz+W+1)+dI(iz+dz+ix+is1)*I(iz+ix+is2)
	  enddo
          !!w1(dz+W+1) = (dz/W)**2;
          w2(dz+W+1) = -exp(-(dz**2)/sigma**2)
	  S=S+corr(dz+W+1)**2
	  S0=S0 + (corr(dz+W+1)-dcorr(dz+W+1))**2
   	enddo
	
	do dz=-W,W
    	  J(nn) = J(nn) + S**(-1)*w2(dz+W+1)*(corr(dz+W+1)**2);
	  J0(nn) = J0(nn) + S0**(-1)*w2(dz+W+1)*(corr(dz+W+1)-dcorr(dz+W+1))**2;
	enddo

	mu1=0.
	do dz=-W,W
    	  mu1(dz+W+1) = mu1(dz+W+1) - 2*S**(-1)*(w2(dz+W+1)-J(nn))*corr(dz+W+1);
	enddo

	hi1=0.
	hi2=0.
	do dz=-W,W
	  do iz=zi(nn)-winz/2,zi(nn)+winz/2
            if (iz+dz .lt. Nz) then
	      adjI(iz+dz+ix+is1) = adjI(iz+dz+ix+is1) + mu1(dz+W+1)*I(iz+ix+is2)
	      adjI(iz+ix+is2)    = adjI(iz+ix+is2)    + mu1(dz+W+1)*I(iz+dz+ix+is1)
	      hi2(iz) = hi2(iz) + mu1(dz+W+1) * I(iz+dz+ix+is1)
	      hi1(iz+dz) = hi1(iz+dz) + mu1(dz+W+1) * I(iz+ix+is2)
	    endif
	  enddo
	enddo
	
	do iz=1,Nz	
	  didi(nn) = didi(nn) - hi1(iz)*dI(iz+ix+is1) - hi2(iz)*dI(iz+ix+is2)
	enddo
	
	totdidi = totdidi + didi(nn)
	Jsum = Jsum + J(nn) - J0(nn)
	
  enddo


  if (myrank==0) write(*,*) 'dIDI = ', totdidi, ' dJ = ',Jsum


  deallocate(corr,dcorr, w2, mu1)





end subroutine power_norm_adjtest


!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!

!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!


subroutine power_norm_adjtest2(J,I,adjI,Nx,Nz,NRHS,nsi,xi,zi,si1,si2)


  use parser
  use fileio 

  implicit none
  
  include 'mpif.h'
  
  integer        :: ix,iz,nn,ixx,dz,inn, is1,is2,myrank,ierr,nprocs
  integer        :: nx, nz, nrhs, nsi, winz, W, ntap
  real           :: I(Nx*Nz*NRHS),adjI(Nx*Nz*NRHS/2), dI(Nx*Nz*NRHS/2), sigma, S,S0,hi1(Nz),hi2(Nz)
  real           :: totdidi, Jsum
  real           :: J(nsi),J0(nsi),didi(nsi),xi(*),zi(*),si1(*),si2(*)
  real, allocatable   :: corr(:),dcorr(:), w2(:),mu1(:),taper(:),ctaper(:)
  character(256) :: Ifile,srcnum
  real  :: pi
    
    pi = 3.1415926
    
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
  
  call frompar('winz',winz,100)
  call frompar('deltaz',W,50)
  call frompar('sigma',sigma,40.)

  allocate(corr(2*W+1),ctaper(2*W+1), dcorr(2*W+1),w2(2*W+1),mu1(2*W+1),taper(winz+1))

  ntap =20
  do iz=1,winz+1
    taper(iz)=1.
  enddo
  do iz=1,2*W+1
    ctaper(iz)=1.
  enddo
  do iz=1,ntap
    taper(iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
    taper(winz+2-iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
    ctaper(iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
    ctaper(2*W+2-iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
  enddo
  
!ifile = 'taperout2'
!call realoutb(ctaper,2*W+1,ifile)
     
 J = 0.
 J0 = 0.
 Jsum = 0.
 do nn=1,nsi
   ix=Nz*(xi(nn)-1)
   is1=Nz*Nx*(si1(nn)-1)
   is2=Nz*Nx*(si2(nn)-1)
   
	corr=0.
	S=0.
	do dz=-W,W
	  do iz=zi(nn)-winz/2,zi(nn)+winz/2
            if (iz+dz .lt. Nz) corr(dz+W+1)=corr(dz+W+1)+I(iz+dz+ix+is1)*taper(iz-zi(nn)+winz/2+1)*I(iz+ix+is2)*taper(iz-zi(nn)+winz/2+1)
	  enddo
          !!w1(dz+W+1) = (dz/W)**2;
          w2(dz+W+1) = -exp(-(dz**2)/sigma**2)
	  S=S+corr(dz+W+1)**2
   	enddo
	do dz=-W,W
	  corr(dz+W+1)=corr(dz+W+1)*ctaper(dz+W+1)
	enddo
	

      
	do dz=-W,W
    	  J(nn) = J(nn) + S**(-1)*w2(dz+W+1)*(corr(dz+W+1)**2);
	enddo

	Jsum = Jsum + J(nn)
  enddo
 !if (myrank==0) write(*,*) 'debug6', Jsum

 do nn=1,nsi
   ix=Nz*(xi(nn)-1)
   is1=Nz*Nx*(si1(nn)-1)
   is2=Nz*Nx*(si2(nn)-1)
   
	corr=0.
	S=0.
	do dz=-W,W
	  do iz=zi(nn)-winz/2,zi(nn)+winz/2
            if (iz+dz .lt. Nz) corr(dz+W+1)=corr(dz+W+1)+I(iz+dz+ix+is1)*taper(iz-zi(nn)+winz/2+1)*I(iz+ix+is2)*taper(iz-zi(nn)+winz/2+1)
	  enddo
          !!w1(dz+W+1) = (dz/W)**2;
          w2(dz+W+1) = -exp(-(dz**2)/sigma**2)
	  S=S+corr(dz+W+1)**2
   	enddo
	do dz=-W,W
	  corr(dz+W+1)=corr(dz+W+1)*ctaper(dz+W+1)
	enddo
	
!ifile = 'corrout'
!call realoutb(corr,2*W+1,ifile)

	mu1=0.
	do dz=-W,W                                          !! Changed to using total error measure here
    	  mu1(dz+W+1) = mu1(dz+W+1) - 2*S**(-1)*(w2(dz+W+1)-Jsum)*corr(dz+W+1);
	enddo
	
 !ifile = 'mu1out'
 !call realoutb(mu1,2*W+1,ifile)
 
	hi1=0.
	hi2=0.
	do dz=-W,W
	  do iz=zi(nn)-winz/2,zi(nn)+winz/2
            if (iz+dz .lt. Nz .and. iz-dz .lt. Nz) then
	      !!adjI(iz+dz+ix+is1) = adjI(iz+dz+ix+is1) + mu1(dz+W+1)*I(iz+ix+is2)
	      adjI(iz+ix+is1) = adjI(iz+ix+is1) + mu1(dz+W+1)*I(iz-dz+ix+is2)*taper(iz-zi(nn)+winz/2+1)	      
	      adjI(iz+ix+is2) = adjI(iz+ix+is2) + mu1(dz+W+1)*I(iz+dz+ix+is1)*taper(iz-zi(nn)+winz/2+1)
	      hi1(iz) = hi1(iz) + mu1(dz+W+1)*I(iz-dz+ix+is2)*taper(iz-zi(nn)+winz/2+1)	      
	      hi2(iz) = hi2(iz) + mu1(dz+W+1)*I(iz+dz+ix+is1)*taper(iz-zi(nn)+winz/2+1)
	    endif
	  enddo
	enddo
	

  enddo




  deallocate(corr,dcorr, w2, mu1)





end subroutine power_norm_adjtest2


!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!

!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!


subroutine power_norm_adjtest3(J,I,adjI,Nx,Nz,NRHS,nsi)


  use parser
  use fileio 

  implicit none
  
  include 'mpif.h'
  
  integer        :: ix,iz,nn,ixx,dz,inn, is1,is2,myrank,ierr,nprocs
  integer        :: nx, nz, nrhs, nsi, winz, W, ntap
  real           :: I(Nx*Nz*NRHS),adjI(Nx*Nz*NRHS/2), dI(Nx*Nz*NRHS/2), sigma, S,S0,hi1(Nz),hi2(Nz)
  real           :: totdidi, Jsum,taperx(15)
  real           :: J(nsi),J0(nsi),didi(nsi),xi(nsi),zi(nsi),si1(nsi),si2(nsi),ziu(nsi),zid(nsi),tmpr(nsi),sigma0(nsi)
  integer        :: W0(nsi),tmp1(nsi),tmp2(nsi)
  real, allocatable   :: corr(:),dcorr(:), w2(:),mu1(:),taper(:),ctaper(:)
  character(256) :: rfile,rfile0,srcnum
  real  :: pi
    
    pi = 3.1415926
    
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
  
  call frompar('scatfile',rfile0)
  !call frompar('deltaz',W,50)
  !call frompar('sigma',sigma,40.)
  
        rfile=trim(rfile0)//'x'
    call realinb(xi,nsi,rfile)
        rfile=trim(rfile0)//'zup'
    call realinb(ziu,nsi,rfile)
        rfile=trim(rfile0)//'zdown'
    call realinb(zid,nsi,rfile)
        rfile=trim(rfile0)//'s1'
    call realinb(si1,nsi,rfile)
        rfile=trim(rfile0)//'s2'
    call realinb(si2,nsi,rfile)
        rfile=trim(rfile0)//'dz'
    call realinb(tmpr,nsi,rfile)
    W0=floor(tmpr)
        rfile=trim(rfile0)//'sigma'
    call realinb(sigma0,nsi,rfile)


 ! xi = xi + 360  !! FOR USING OLD ERROR FUNCTIONS WITH WIDE MODEL



  !ziu=ziu-10
  !zid=zid+10


 !%%%%%%%%%%%%%%%%%
 !W0=15
 !%%%%%%%%%%%%%%%%%

       tmp1=2*W0+10
       allocate(corr(maxval(tmp1,nsi)),ctaper(maxval(tmp1,nsi)), dcorr(maxval(tmp1,nsi)))
	allocate(w2(maxval(tmp1,nsi)),mu1(maxval(tmp1,nsi)))
!if (myrank==0) write(*,*) 'debug6',maxval(tmp1,nsi)
	tmp2=floor(zid-ziu)+10
	allocate(taper(maxval(tmp2,nsi)))
!if (myrank==0) write(*,*) 'debug6',maxval(tmp2,nsi)


 J = 0.
 J0 = 0.
 Jsum = 0.
 
 
 do nn=1,nsi

  if (si1(nn) .ne. 29 .and. si2(nn) .ne. 29) then

  W=W0(nn)
  sigma=sigma0(nn)
  winz=2*floor((zid(nn)-ziu(nn))/2)
  zi(nn)=floor((zid(nn)+ziu(nn))/2)


! if (myrank==0) write(*,*) 'debug6', nn, W, sigma, winz, zi(nn)


  ntap = 15
  taper = 0.
  ctaper = 0.
  do iz=1,winz+1
    taper(iz)=1.
  enddo
  do iz=1,2*W+1
    ctaper(iz)=1.
  enddo
  do iz=1,ntap
    taper(iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
    taper(winz+2-iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
    ctaper(iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
    ctaper(2*W+2-iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
  enddo


   ix=Nz*(xi(nn)-1)
   is1=Nz*Nx*(si1(nn)-1)
   is2=Nz*Nx*(si2(nn)-1)
   
	corr=0.
	S=0.
	do dz=-W,W
	  do iz=zi(nn)-winz/2,zi(nn)+winz/2
            if (iz+dz .lt. Nz) corr(dz+W+1)=corr(dz+W+1)+I(iz+dz+ix+is1)*taper(iz-zi(nn)+winz/2+1)*I(iz+ix+is2)*taper(iz-zi(nn)+winz/2+1)
	  enddo
          !!w1(dz+W+1) = (dz/W)**2;
          w2(dz+W+1) = -exp(-(dz**2)/sigma**2)
	  S=S+corr(dz+W+1)**2
   	enddo
	do dz=-W,W
	  corr(dz+W+1)=corr(dz+W+1)*ctaper(dz+W+1)
	  
	enddo
	
      
	do dz=-W,W
    	  J(nn) = J(nn) + S**(-1)*w2(dz+W+1)*(corr(dz+W+1)**2);
	enddo

	Jsum = Jsum + J(nn)
		  !if (myrank==0) write(*,*) 'debug7b',W,sigma,winz,J(nn)
		
  endif	
  enddo
 !if (myrank==0) write(*,*) 'debug6', Jsum


  adjI=0.

  do nn=1,nsi
  
  if (si1(nn) .ne. 29 .and. si2(nn) .ne. 29) then

  W=W0(nn)
  sigma=sigma0(nn)
  winz=2*floor((zid(nn)-ziu(nn))/2)
  zi(nn)=floor((zid(nn)+ziu(nn))/2)
  
 !if (myrank==0) write(*,*) 'debug7', nn, W, sigma, winz, zi(nn)
 
  ntap = 15
  taper = 0.
  ctaper = 0.
  do iz=1,winz+1
    taper(iz)=1.
  enddo
  do iz=1,2*W+1
    ctaper(iz)=1.
  enddo
  do iz=1,ntap
    taper(iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
    taper(winz+2-iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
    ctaper(iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
    ctaper(2*W+2-iz) = sin((real(iz)-1.)/(real(ntap)-1.)*PI/2)**2
  enddo

   ix=Nz*(xi(nn)-1)
   is1=Nz*Nx*(si1(nn)-1)
   is2=Nz*Nx*(si2(nn)-1)
   
	corr=0.
	S=0.
	do dz=-W,W
	  do iz=zi(nn)-winz/2,zi(nn)+winz/2
            if (iz+dz .lt. Nz) corr(dz+W+1)=corr(dz+W+1)+I(iz+dz+ix+is1)*taper(iz-zi(nn)+winz/2+1)*I(iz+ix+is2)*taper(iz-zi(nn)+winz/2+1)
	    !if (iz+dz .lt. Nz) corr(dz+W+1)=corr(dz+W+1)+I(iz+dz+ix+is1)*I(iz+ix+is2)
	  enddo
          !!w1(dz+W+1) = (dz/W)**2;
          w2(dz+W+1) = -exp(-(dz**2)/sigma**2)
	  S=S+corr(dz+W+1)**2
	  !if (myrank==0) write(*,*) 'debug7a',dz,dz+W+1,w2(dz+W+1),corr(dz+W+1)
   	enddo
	do dz=-W,W
	  corr(dz+W+1)=corr(dz+W+1)*ctaper(dz+W+1)
	enddo
	
 !rfile = 'corrout'
 !call realoutb(w2,maxval(tmp1,nsi),rfile)
 !rfile = 'taperout'
 !call realoutb(taper,maxval(tmp2,nsi),rfile)
  !rfile = 'ctaperout'
! call realoutb(ctaper,maxval(tmp1,nsi),rfile)

	mu1=0.
	do dz=-W,W                                          !! Changed to using total error measure here
    	  !mu1(dz+W+1) = mu1(dz+W+1) - 2*S**(-1)*(w2(dz+W+1)-J(nn))*corr(dz+W+1);
	  mu1(dz+W+1) = mu1(dz+W+1) - 2*S**(-1)*(w2(dz+W+1)-J(nn))*corr(dz+W+1);
  
	  	  !if (myrank==0) write(*,*) 'debug7a',dz,dz+W+1,corr(dz+W+1),mu1(dz+W+1),Jsum
	enddo
	
 !rfile = 'mu1out'
 !call realoutb(mu1,maxval(tmp1,nsi),rfile)
 
	hi1=0.
	hi2=0.
	do dz=-W,W
	  do iz=zi(nn)-winz/2,zi(nn)+winz/2
            if (iz+dz .lt. Nz .and. iz-dz .lt. Nz) then
	      !!adjI(iz+dz+ix+is1) = adjI(iz+dz+ix+is1) + mu1(dz+W+1)*I(iz+ix+is2)
	      adjI(iz+ix+is1) = adjI(iz+ix+is1) + mu1(dz+W+1)*I(iz-dz+ix+is2)*taper(iz-zi(nn)+winz/2+1)	      
	      adjI(iz+ix+is2) = adjI(iz+ix+is2) + mu1(dz+W+1)*I(iz+dz+ix+is1)*taper(iz-zi(nn)+winz/2+1)
	      hi1(iz) = hi1(iz) + mu1(dz+W+1)*I(iz-dz+ix+is2)*taper(iz-zi(nn)+winz/2+1)	      
	      hi2(iz) = hi2(iz) + mu1(dz+W+1)*I(iz+dz+ix+is1)*taper(iz-zi(nn)+winz/2+1)
	    endif
	  enddo
	enddo
	
  endif

  enddo

  deallocate(corr,dcorr, w2, mu1)

  do ix=1,15
    taperx(ix) = exp(-real(ix-8)**2/2.8**2)
  enddo
  
  
  dI=0.
  do is2=60,Nx-60
    do is1=1,nrhs/2
      do ix=1,15
        do iz=1,nz
	  ixx = Nz*(is2 + (ix-8) - 1)
	  dI(iz+ixx+Nz*Nx*(is1-1)) = dI(iz+ixx+Nz*Nx*(is1-1))+adjI(iz+Nz*(is2-1)+Nz*Nx*(is1-1))*taperx(ix)
	enddo
      enddo
    enddo
  enddo


  adjI=dI


end subroutine power_norm_adjtest3

!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!





subroutine power_norm2D(J,I,adjI,Nx,Nz,NRHS,nsi,xi,zi,si1,si2)


  use parser
  use fileio 

  implicit none
  
  include 'mpif.h'
  
  integer        :: ix,iz,nn,ixx,dz,dx, is1,is2,myrank,ierr,nprocs
  integer        :: nx, nz, nrhs, nsi, winz,winx, Wz,Wx
  real           :: I(Nx*Nz*NRHS),adjI(Nx*Nz*NRHS/2),tmpI(Nx*Nz*NRHS/2), sigma, S
  real           :: J(nsi),tmpJ(nsi),xi(*),zi(*),si1(*),si2(*)
  real, allocatable   :: corr(:,:),w2(:,:),mu1(:,:)
  !character(256) :: Ifile,srcnum
  
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
 call MPI_Comm_size( MPI_COMM_WORLD, nprocs, ierr )
   
   
  call frompar('winz',winz,100)
  call frompar('winx',winx,100)
  call frompar('deltaz',Wz,50)
  call frompar('deltax',Wx,50)
  call frompar('sigma',sigma,40.)

  allocate(corr(2*Wz+1,2*Wx+1),w2(2*Wz+1,2*Wx+1),mu1(2*Wz+1,2*Wx+1))


 J = 0.
 adjI= 0.
 tmpI=0.
 tmpJ=0.
 

 
 
 !do nn=1,nsi
 do nn=1+myrank,nsi,nprocs
   ix=Nz*(xi(nn)-1)
   is1=Nz*Nx*(si1(nn)-1)
   is2=Nz*Nx*(si2(nn)-1)
   
	corr=0.
	S=0.
	do dz=-Wz,Wz
	  do dx=-Wx,Wx
	    do iz=zi(nn)-winz/2,zi(nn)+winz/2
	     do ix=xi(nn)-winx/2,xi(nn)+winx/2
	     ixx=Nz*(ix-1)
              if (iz+dz.lt.Nz .and. ix+dx.lt.Nx .and. ix-dx.gt.1 .and. sqrt(real(dz**2+dx**2)).lt.30. ) then
	        corr(dz+Wz+1,dx+Wx+1)=corr(dz+Wz+1,dx+Wx+1)+I(iz+dz+ixx+Nz*dx+is1)*I(iz+ixx+is2)
	      endif
             enddo
	    enddo
            w2(dz+Wz+1,dx+Wz+1) = -exp(-(dz**2 + dx**2)/sigma**2)
	    S=S+corr(dz+Wz+1,dx+Wx+1)**2
	  enddo 
   	enddo
	
	do dz=-Wz,Wz
	  do dx=-Wx,Wx
    	    tmpJ(nn) = tmpJ(nn) + S**(-1)*w2(dz+Wz+1,dx+Wx+1)*(corr(dz+Wz+1,dx+Wx+1)**2);
	  enddo
	enddo

	mu1=0.
	do dz=-Wz,Wz
	  do dx=-Wx,Wx
    	    mu1(dz+Wz+1,dx+Wx+1) = mu1(dz+Wz+1,dx+Wx+1) - 2*S**(-1)*(w2(dz+Wz+1,dx+Wx+1)-J(nn))*corr(dz+Wz+1,dx+Wx+1);
	  enddo
	enddo

	do dz=-Wz,Wz
	  do dx=-Wx,Wx
	    do iz=zi(nn)-winz/2,zi(nn)+winz/2
	      do ix=xi(nn)-winx/2,xi(nn)+winx/2
	        ixx=Nz*(ix-1)
                if (iz+dz.lt.Nz .and. ix+dx.lt.Nx .and. ix-dx.gt.1 .and. sqrt(real(dz**2+dx**2)).lt.30. ) then
	          tmpI(iz+dz+ixx+Nz*dx+is1) = tmpI(iz+dz+ixx+Nz*dx+is1) + mu1(dz+Wz+1,dx+Wx+1)*I(iz+ixx+is2)
	          tmpI(iz+ixx+is2)    = tmpI(iz+ixx+is2)    + mu1(dz+Wz+1,dx+Wx+1)*I(iz+dz+ixx+Nz*dx+is1)
	        endif
	      enddo
	    enddo
	  enddo
	enddo
	
	if (myrank==0) write(*,*) 'Source pair ', nn, tmpJ(nn)!!,S,si1(nn),si2(nn),xi(nn),zi(nn)
  enddo


  call MPI_ALLREDUCE(tmpI,adjI,Nx*Nz*NRHS/2,MPI_REAL,MPI_SUM,MPI_COMM_WORLD,ierr)
  call MPI_ALLREDUCE(tmpJ,J,NSI,MPI_REAL,MPI_SUM,MPI_COMM_WORLD,ierr)


  deallocate(corr, w2, mu1)



end subroutine power_norm2D


!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!

subroutine power_norm2D_test(J,I,adjI,winx,winz,Wx,Wz,sigmax,sigmaz,Nx,Nz,NRHS,nsi,xi,zi,si1,si2)


  use parser
  use fileio 

  implicit none
  
  include 'mpif.h'
  
  integer        :: ix,iz,nn,ixx,dz,dx, is1,is2,myrank,ierr,nprocs
  integer        :: nx, nz, nrhs, nsi, winz,winx, Wz,Wx
  real           :: I(Nx*Nz*NRHS),adjI(Nx*Nz*NRHS/2),tmpI(Nx*Nz*NRHS/2), sigmax, sigmaz, S
  real           :: J(nsi),tmpJ(nsi),xi(*),zi(*),si1(*),si2(*)
  real, allocatable   :: corr(:,:),w2(:,:),mu1(:,:)
  !character(256) :: Ifile,srcnum
  
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
 call MPI_Comm_size( MPI_COMM_WORLD, nprocs, ierr )
   
  if (.false.) then
  call frompar('winz',winz,100)
  call frompar('winx',winx,100)
  call frompar('deltaz',Wz,50)
  call frompar('deltax',Wx,50)
  call frompar('sigma',sigmaz,40.)
  call frompar('sigma',sigmax,40.)
  endif
  
  allocate(corr(2*Wz+1,2*Wx+1),w2(2*Wz+1,2*Wx+1),mu1(2*Wz+1,2*Wx+1))


  !if (myrank==0) write(*,*) winx,winz,Wx,Wz,sigmax,sigmaz,Nx,Nz,NRHS

 J = 0.

 adjI= 0.
 tmpI=0.
 tmpJ=0.
 
 
 
 !do nn=1,nsi
 do nn=1+myrank,nsi,nprocs
   ix=Nz*(xi(nn)-1)
   is1=Nz*Nx*(si1(nn)-1)
   is2=Nz*Nx*(si2(nn)-1)
   
	corr=0.
	S=0.
	do dz=-Wz,Wz
	  do dx=-Wx,Wx
	    do iz=zi(nn)-winz/2,zi(nn)+winz/2
	     do ix=xi(nn)-winx/2,xi(nn)+winx/2
	     ixx=Nz*(ix-1)
              if (iz+dz.lt.Nz .and. ix+dx.lt.Nx .and. ix-dx.gt.1 .and. sqrt(real(dz**2+dx**2)).lt.Wx ) then
	        corr(dz+Wz+1,dx+Wx+1)=corr(dz+Wz+1,dx+Wx+1)+I(iz+dz+ixx+Nz*dx+is1)*I(iz+ixx+is2)
	      endif
             enddo
	    enddo
            w2(dz+Wz+1,dx+Wz+1) = -exp(-(dz**2)/sigmaz**2 -(dx**2)/sigmax**2 )
	    
	    S=S+corr(dz+Wz+1,dx+Wx+1)**2
	  enddo 
   	enddo
	
	do dz=-Wz,Wz
	  do dx=-Wx,Wx
    	    tmpJ(nn) = tmpJ(nn) + S**(-1)*w2(dz+Wz+1,dx+Wx+1)*(corr(dz+Wz+1,dx+Wx+1)**2);
	  enddo
	enddo

	mu1=0.
	do dz=-Wz,Wz
	  do dx=-Wx,Wx
    	    !mu1(dz+Wz+1,dx+Wx+1) = mu1(dz+Wz+1,dx+Wx+1) - 2*S**(-1)*(w2(dz+Wz+1,dx+Wx+1)-J(nn))*corr(dz+Wz+1,dx+Wx+1);
	  enddo
	enddo

	do dz=-Wz,Wz
	  do dx=-Wx,Wx
	    do iz=zi(nn)-winz/2,zi(nn)+winz/2
	      do ix=xi(nn)-winx/2,xi(nn)+winx/2
	        ixx=Nz*(ix-1)
                if (iz+dz.lt.Nz .and. ix+dx.lt.Nx .and. ix-dx.gt.1 .and. sqrt(real(dz**2+dx**2)).lt.Wx ) then
	          !tmpI(iz+dz+ixx+Nz*dx+is1) = tmpI(iz+dz+ixx+Nz*dx+is1) + mu1(dz+Wz+1,dx+Wx+1)*I(iz+ixx+is2)
	          !tmpI(iz+ixx+is2)    = tmpI(iz+ixx+is2)    + mu1(dz+Wz+1,dx+Wx+1)*I(iz+dz+ixx+Nz*dx+is1)
	        endif
	      enddo
	    enddo
	  enddo
	enddo
	
	!if (myrank==0) write(*,*) 'Source pair ', nn, tmpJ(nn),S,si1(nn),si2(nn),xi(nn),zi(nn)
  enddo


  call MPI_ALLREDUCE(tmpI,adjI,Nx*Nz*NRHS/2,MPI_REAL,MPI_SUM,MPI_COMM_WORLD,ierr)
  call MPI_ALLREDUCE(tmpJ,J,NSI,MPI_REAL,MPI_SUM,MPI_COMM_WORLD,ierr)



  deallocate(corr, w2, mu1)



end subroutine power_norm2D_test


!!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!

subroutine find_scatters(I, Nx, Nz, NRHS, nsi, xi,zi,si1,si2)


  use parser
  use fileio 

  implicit none
  
  include 'mpif.h'
  
  integer        :: ix,iz,nn,ixx,dz, is1, is2,myrank,ierr,ii2
  integer        :: nx, nz, nrhs, nsi, winz, W, nscats
  real           :: I(Nx*Nz*NRHS), thresh, corr,fudg(18)
  real           :: J(nsi),xi(*),zi(*),xi0(1000),zi0(1000),si1(*),si2(*),scats(Nx*Nz)
  character(256) :: scatfile

    
  call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
 
  call frompar('winz',winz,100)
  call frompar('threshold',thresh,5.E-14)
  call frompar('scatfile',scatfile)
  
  winz=100

  call realinb(scats,Nx*Nz,scatfile)

  nscats=0
  do ix=1,Nx
    do iz=1,Nz
        ixx=Nz*(ix-1)
        if (scats(iz+ixx) .lt. 254.) then
	  nscats = nscats + 1
	  xi0(nscats) = ix
	  zi0(nscats) = iz
	endif
    enddo
  enddo

if (myrank==0) write(*,*) 'Number of scatterers: ',nscats

fudg=(/1, 2, 3, 4, 5, 6, 7, 8, 9, 21, 22, 23, 24, 25, 26, 27, 28, 29/)

nsi = 0

 !do is1=14,14 
 !   do is2=20,20 
  !    do nn= 42,42 
 do is1=1,NRHS/2,2 !!1,18  !!
    do is2=2,NRHS/2,2  !!10,20 !!
      do nn=1,nscats   
      
        !is2=fudg(ii2)
        ix=Nz*(xi0(nn)-1)
        corr=0

        do iz=zi0(nn)-winz/2,zi0(nn)+winz/2
          corr=corr+I(iz+ix+Nx*Nz*(is1-1))*I(iz+ix+Nx*Nz*(is2-1))
        enddo
        if (corr .gt. thresh) then
	  nsi=nsi+1
          si1(nsi) = is1
          si2(nsi) = is2
          xi(nsi) = xi0(nn)
          zi(nsi) = zi0(nn)
	  !!if (myrank==0) write(*,*) si1(nsi),si2(nsi),xi(nsi),zi(nsi)
        endif
      enddo
    enddo
   
enddo
if (myrank==0) write(*,*) 'Number of pairs: ',nsi


end subroutine find_scatters

!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
subroutine smootherf1D(adjI,Nx,Nz,Nrhs)

  use parser
  use fileio 

  implicit none
  
  include 'mpif.h'
  
  integer        :: ix,iz,nn,ixx,izz, is1,is2,myrank,ierr,nprocs
  integer        :: nx, nz, nrhs, nsi, winz, winx, nscats
  real           :: I(Nx*Nz*NRHS),adjI(Nx*Nz*NRHS/2),tmpI(Nx*Nz*NRHS/2)
  real           :: xi0(1000),zi0(1000),taperx(15),scats(Nx*Nz)
  real,allocatable :: taperz(:)
  character(256) :: scatfile
  
  call frompar('winz',winz,100)
  call frompar('winx',winx,100)
  call frompar('scatfile',scatfile)
  call realinb(scats,Nx*Nz,scatfile)

  allocate(taperz(winz+14))

  nscats=0
  do ix=1,Nx
    do iz=1,Nz
        ixx=Nz*(ix-1)
        if (scats(iz+ixx) .lt. 254.) then
	  nscats = nscats + 1
	  xi0(nscats) = ix
	  zi0(nscats) = iz
	endif
    enddo
  enddo
  
  
  taperz=0.
  
  do ix=1,15
    taperx(ix) = exp(-real(ix-8)**2/2.8**2)
  enddo
  do iz=1,winz+14
    do ix=8,winz+7
      taperz(iz)=taperz(iz)+ exp(-real(ix-iz)**2/2.8**2)
    enddo
  enddo
  taperz=taperz/maxval(taperz,winz+14)
  
  
  do nn=1,nscats
    do is1=1,nrhs/2
      do ix=1,15
        do iz=1,winz+14
	  ixx = Nz*(xi0(nn) + (ix-8) - 1)
	  izz = zi0(nn) + (iz-winz/2-7)
	  !!!tmpI(izz+ixx+Nz*Nx*(is1-1)) = tmpI(izz+ixx+Nz*Nx*(is1-1))+adjI(izz+Nz*(xi0(nn)-1)+Nz*Nx*(is1-1))*taperx(ix)
	  !*taperx(ix)!*taperz(iz)
	enddo
      enddo
    enddo
  enddo
    
  do nn=1,nscats
    do is1=1,nrhs/2
      do ix=1,15
        do iz=1,nz
	  ixx = Nz*(xi0(nn) + (ix-8) - 1)
	  izz = iz
	  tmpI(izz+ixx+Nz*Nx*(is1-1)) = tmpI(izz+ixx+Nz*Nx*(is1-1))+adjI(izz+Nz*(xi0(nn)-1)+Nz*Nx*(is1-1))*taperx(ix)
	enddo
      enddo
    enddo
  enddo
  
  
 adjI=tmpI

  deallocate(taperz)



end subroutine smootherf1D


!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!
!!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!





end module powernorm
